#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
AKR-Agent Simple Example
"""

import asyncio
import json
import os
import sys

# Add project root directory to Python path
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from akr_agent.agent import Agent
from akr_agent.tools.base import ToolCenter
from akr_agent.tools.tool_llm import LLMCallTool
from akr_agent.tools.tool_search import DuckDuckGoSearchTool


from loguru import logger

async def main():
    """Main function"""
    # Register tools
    ToolCenter.register(
        tool=LLMCallTool(
            api_key=os.environ.get("DASHSCOPE_API_KEY"),
            base_url="https://dashscope.aliyuncs.com/compatible-mode/v1",
            model="qwen-turbo-latest",
            temperature=0.7,
            max_tokens=1000,
            stream=True,
        )
    )
    ToolCenter.register(tool=DuckDuckGoSearchTool(), name="DuckDuckGoSearchTool")
    
    # Create Agent instance
    agent = Agent(config_dir="examples/prompts/CoachLi/v1", sid="test")
    
    # User input
    user_input = "具体描述食物组成、克重预估"
    image_url = "https://i.ibb.co/NMhXwPD/72d59001-b6a7-473c-9b76-32ea9f959363.png"
    logger.info(f"\n--- User Input ---\n{user_input}")
    
    # Run Agent and get response
    logger.info("\n--- Agent Response ---")
    async for chunk in agent.run_dynamic(user_input, image_url=image_url, image_url_detail="high"):
        logger.info(chunk.content)
    
    logger.info("\n\n--- Done ---\n")
    logger.info("\n--- All Context ---\n")
    logger.info(
        json.dumps(agent._context_manager.get_context().to_dict(), indent=2, ensure_ascii=False)
    )


if __name__ == "__main__":
    asyncio.run(main())



import os
from typing import Optional
from loguru import logger
import traceback

from typing import Any, AsyncGenerator, Dict, List
from jinja2 import Environment, select_autoescape, TemplateError

# Setup Jinja2 environment
# You might want to move this to a more central place if used elsewhere
jinja_env = Environment(
    loader=None,  # We'll load templates from strings
    autoescape=select_autoescape(["html", "xml"]),  # Basic autoescaping
)

from ..context_manager import ObservableCtx
from ..rule_config import RuleConfig
from ..llm.openai_client import OpenAIClient
from .base import Tool, ToolCenter


class LLMCallTool(Tool):
    """
    LLM Call Tool, used to call LLM

    Supported parameters:
    - system_prompt: System prompt
    - prompt: Prompt
    - prompt_detail: Prompt detail
    - ctx: Observable context
    - rule_config: Rule config
    - extra: Extra config
        - tools: Tool list
    """

    def __init__(self, llm_client: Optional[OpenAIClient] = None, **kwargs):
        self.llm_client = llm_client
        if not self.llm_client:
            self.llm_client = OpenAIClient(
                model=kwargs.get("model", "gpt4o-mini"),
                api_key=kwargs.get("api_key", os.environ.get("OPENAI_API_KEY")),
                base_url=kwargs.get("base_url", "https://api.openai.com"),
                temperature=kwargs.get("temperature", 0.7),
                max_tokens=kwargs.get("max_tokens", 1024),
                stream=kwargs.get("stream", True),
            )

    async def run(
        self,
        user_input: str,
        system_prompt: str,
        prompt: str,
        prompt_detail: str,
        **kwargs,
    ) -> AsyncGenerator[str, None]:
        """Execute LLM call

        Args:
            user_input: User input, the content of user input when calling LLM
            system_prompt: The most important system prompt, defining the goal, requirements, constraints, and expected output
            prompt: Prompt, used to supplement system prompt, strengthening key points
            prompt_detail: Detailed explanation, guidance, etc. for key points, encouraging the model to perform inference steps, add examples, or standardize output format before reaching a conclusion
            **kwargs: Other parameters
        """
        system_prompt = await self._render_prompt(
            system_prompt=system_prompt,
            prompt=prompt,
            prompt_detail=prompt_detail,
            **kwargs,
        )

        logger.debug(f"llm tool call, kwargs={kwargs}")

        tool_defs = await self._build_tool_defs(**kwargs)

        async for chunk in self.llm_client.invoke_stream(
            system_prompt=system_prompt,
            user_input=user_input,
            messages=[],
            run_tool_func=ToolCenter.run_tool,
            tools=tool_defs,
            ctx_manager=kwargs.get("ctx_manager"),
            image_url=kwargs.get("image_url"),
        ):
            yield chunk

    async def _render_prompt(
        self, system_prompt: str, prompt: str, prompt_detail: str, **kwargs
    ) -> str:
        # Ensure system_prompt is not None
        if system_prompt is None:
            system_prompt = ""

        if not (prompt or prompt_detail):
            return system_prompt

        if prompt:
            system_prompt = system_prompt + "\n\n" + prompt
        if prompt_detail:
            system_prompt += "\n\n" + prompt_detail

        ctx: ObservableCtx = kwargs.get("ctx")
        rule_config: RuleConfig = kwargs.get("rule_config")

        if not ctx:
            return system_prompt

        custom_render_ctx = (
            {k: ctx.get(k) for k in rule_config.depend_ctx_key}
            if rule_config.depend_ctx_key
            else ctx.to_dict()
        )
        try:
            template = jinja_env.from_string(system_prompt)
            system_prompt = template.render(**custom_render_ctx)
        except TemplateError as e:
            logger.error(f"Template rendering error: {e}, trace={traceback.format_exc()}")
            # Downgrade strategy: return unrendered prompt
            system_prompt = system_prompt + "\n\n[Note: Template rendering failed]"
        except Exception as e:
            logger.error(
                f"Error rendering prompt template for rule: {rule_config.name}: {e}, trace={traceback.format_exc()}"
            )

        return system_prompt

    async def _build_tool_defs(self, **kwargs) -> List[Dict[str, Any]]:
        tool_defs = []
        config_tool_names: List[str] = kwargs.get("tools", [])
        logger.debug(f"tool exec depends tools={config_tool_names}")
        if not config_tool_names or len(config_tool_names) == 0:
            return tool_defs
        for tool_name in config_tool_names:
            tool_def = ToolCenter.get_definition(name=tool_name)
            if tool_def:
                tool_defs.append({"type": "function", "function": tool_def})
                logger.debug(f"Use tool: {tool_name} {tool_defs}")
            else:
                logger.warning(f"Cannot get tool: {tool_name}, def={tool_def}")

        return tool_defs



from akr_agent.libs.code.code_exec import PythonCodeExecutor
from typing import Union, List


def realtime_web_search(query: Union[str, List[str]]) -> str:
    """Searches the web for the given query and returns the results."""
    from akr_agent.libs.search.tavily import TavilySearchAPIWrapper
    import os
    import json
    wrapper = TavilySearchAPIWrapper(tavily_api_key=os.environ.get("TAVILY_API_KEY", None))
    if isinstance(query, list):
        results = [wrapper.results(q, max_results=2) for q in query]
    else:
        results = wrapper.results(query, max_results=3)
    results = [result for result in results if result.get("content")]
    final_result = json.dumps(results, indent=2, ensure_ascii=False)

    return final_result


def web_crawler_to_markdown(url: Union[str, List[str]]) -> str:
    """Crawls the given URL and returns the content as Markdown."""
    import asyncio
    from akr_agent.libs.crawler.crawl4ai import Crawl4AIWrapper
    wrapper = Crawl4AIWrapper()
    if isinstance(url, list):
        results = [asyncio.run(wrapper.crawl(u)) for u in url]
        result = "\n\n".join(results)
    else:
        result = asyncio.run(wrapper.crawl(url))
    
    return result



executor = PythonCodeExecutor()
code_text = """
import json
def search_best_llm():
    try:
        leaderboard_info = realtime_web_search(query="Hugging Face Open LLM Leaderboard current top models")
        recent_reviews = realtime_web_search(query="best open source LLM May 2024 review OR Llama 3 vs Mixtral vs Qwen")
        return {
            "leaderboard_info": leaderboard_info,
            "recent_reviews": recent_reviews
        }
    except Exception as e:
        return {"error": str(e)}

search_results = search_best_llm()
search_results
"""
result = executor.run(code_text, realtime_web_search=realtime_web_search, web_crawler_to_markdown=web_crawler_to_markdown)

print("--==--" * 10)
print("\n\n")
for k, v in result.items():
    print(f"{k}: {v}")






SYSTEM_PROMPT = """
You are the best coding assistant assigned with the task of problem-solving.
Follow a structured approach to problem-solving by outlining your plan, implementing it in code, executing it, and presenting the final solution.

# Steps

1. **Thinking Process:** 
   - Provide a complete, step-by-step outline of your plan to solve the provided problem.
   - Detail all the necessary steps you will take, considering constraints.

2. **Code Implementation:** 
    - Based on your detailed thinking process, write the complete Python code required to execute the plan and work towards the solution.
3. **Execution Block:** 
    - Enclose the complete code you wrote in Step 2 within <execute>...</execute> tags.
    - Make sure the last line of the code is the variable name you want to return.
    - *(Note: After you provide the thinking process and the code in the <execute> block, the environment will execute the code and provide its output back to you.)*
4. **Execution Result Analysis:**    
    - Analyze the output from the execution environment. If result solve the problem, Provide the final answer or required output within <solution>...</solution> tags.
    - If result does not solve the problem, analyze the error and step back to the thinking process and code implementation.
5. **Final Solution:** 
    - After reviewing the output from the execution environment, process the results and provide the final answer or required output within <solution>...</solution> tags.


# Constraints and Usage Rules:

* You have only one opportunity to interact with the execution environment using the `<execute>` tag.
* The `<solution>` tag should be used only once at the end of the process to deliver the final result based on the code execution output.
* You python code should only use the tools provided by the environment and built-in python packages.


# Available Tools:

* realtime_web_search(query_list: List[str]) -> str
> This tool provides access to real-time online information. Use it to retrieve the most recent information you need.

* web_crawler_to_markdown(url_list: List[str]) -> str
> This tool allows you to crawl multiple urls and retrieve their content. Use it to extract information from multiple websites.
"""


from typing import Dict, Set, Optional
import traceback
import uuid
import loguru
from loguru._logger import Logger
from collections import defaultdict

from .observable_ctx import ObservableCtx
from .rule_config import RuleConfig


logger = loguru.logger

class RuleRegistry:
    """
    Rule registry, responsible for rule registration, indexing and query
    """
    
    def __init__(self):
        self._rules: Dict[str, RuleConfig] = {}  # rule_id -> RuleConfig
        self._key_to_rules = defaultdict(set)  # ctx_key -> {rule_ids}
        self._rule_priorities: Dict[str, int] = {}  # rule_id -> priority
    
    def generate_rule_id(self, rule_name: str) -> str:
        """Generate unique rule ID"""
        return f"{rule_name}_{str(uuid.uuid4())[:8]}"
    
    def register_rule(self, rule_config: RuleConfig) -> str:
        """Register new rule"""
        rule_id = self.generate_rule_id(rule_config.name)
        self._rules[rule_id] = rule_config
        
        # Set priority
        priority = getattr(rule_config, "priority", 0)
        self._rule_priorities[rule_id] = priority
        
        # Add to index
        for key in rule_config.depend_ctx_key:
            self._key_to_rules[key].add(rule_id)
        
        logger.debug(
            f"Registered rule: {rule_id} (Name: {rule_config.name}, "
            f"Depend_ctx_key: {rule_config.depend_ctx_key}, "
            f"Condition: {rule_config.match_condition}, Priority: {priority})"
        )
        
        return rule_id
    
    def unregister_rule(self, rule_id: str) -> None:
        """Unregister rule"""
        if rule_id not in self._rules:
            return
            
        rule_config = self._rules[rule_id]
        
        # Remove from index
        for key in rule_config.depend_ctx_key:
            if rule_id in self._key_to_rules[key]:
                self._key_to_rules[key].remove(rule_id)
        
        # Remove rule config and priority
        del self._rules[rule_id]
        if rule_id in self._rule_priorities:
            del self._rule_priorities[rule_id]
        
        logger.debug(f"Unregistered rule: {rule_id}")
    
    def get_rule(self, rule_id: str) -> Optional[RuleConfig]:
        """Get rule config"""
        return self._rules.get(rule_id)
    
    def get_rules_for_key(self, key: str) -> Set[str]:
        """Get all rule IDs that depend on the specified key"""
        return self._key_to_rules.get(key, set())
    
    def get_rule_priority(self, rule_id: str) -> int:
        """Get rule priority"""
        return self._rule_priorities.get(rule_id, 0)
    
    def check_rule_condition(self, rule_id: str, ctx: ObservableCtx) -> bool:
        """Check if the rule condition is satisfied"""
        rule_config = self.get_rule(rule_id)
        if not rule_config:
            return False
            
        condition = rule_config.match_condition
        if not condition:
            return True
            
        try:
            # Prepare safe execution environment
            eval_globals = {
                "__builtins__": {
                    "True": True,
                    "False": False,
                    "None": None,
                    "str": str,
                    "int": int,
                    "float": float,
                    "bool": bool,
                    "list": list,
                    "dict": dict,
                    "set": set,
                    "tuple": tuple,
                    "len": len,
                    "Exception": Exception,
                },
                "ctx": ctx,
            }
            result = eval(condition, eval_globals, {})
            logger.debug(
                f"Condition '{condition}' for rule {rule_id} evaluated to: {result}"
            )
            return bool(result)
        except Exception as e:
            logger.error(f"Error evaluating condition for rule {rule_id}: {e} {traceback.format_exc()}")
            return False


from .base import Tool


class DocIndexTool(Tool):
    """
    Document Index Tool, used to index document content into vector database
    """

    def __init__(self, **kwargs):
        pass

    def run(self, **kwargs):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Prompt configuration engine, responsible for loading and managing Agent prompt configuration
"""
from loguru import logger
import os
from typing import List, Dict, Any
import yaml
from jinja2 import Template

from .rule_config import AgentConfig, AgentMeta, RuleConfig


class AgentConfigEngine:
    """Agent configuration engine, responsible for loading and building AgentConfig from YAML files"""

    @staticmethod
    def _load_yaml(file_path: str) -> Dict[str, Any]:
        """Load YAML file"""
        with open(file_path, "r", encoding="utf-8") as f:
            return yaml.safe_load(f)

    @staticmethod
    def _render_template(template_str: str, context: Dict[str, Any]) -> str:
        """Render Jinja2 template"""
        template:Template = Template(template_str)
        return template.render(**context)

    @staticmethod
    def _load_system_prompt(base_config_path: str, meta_data: Dict[str, Any]) -> str:
        """Load and render system prompt"""
        system_prompt_path = os.path.join(base_config_path, "system_prompt.yaml")
        if not os.path.exists(system_prompt_path):
            raise FileNotFoundError(
                f"System prompt file not found: {system_prompt_path}"
            )

        system_prompt_data = AgentConfigEngine._load_yaml(system_prompt_path)
        template_str = system_prompt_data.get("content", "")

        # Render template
        context = {
            "meta": meta_data.get("meta", {}),
            "agent": meta_data.get("agent", {}),
        }

        return AgentConfigEngine._render_template(template_str, context)

    @staticmethod
    def _load_rules(base_config_path: str) -> List[RuleConfig]:
        """Load all rule configurations"""
        rules_dir = os.path.join(base_config_path, "rules")
        if not os.path.isdir(rules_dir):
            raise NotADirectoryError(f"Rules directory not found: {rules_dir}")

        rules = []
        for rule_file in os.listdir(rules_dir):
            if not rule_file.endswith(".yaml"):
                continue

            rule_path = os.path.join(rules_dir, rule_file)
            rule_data = AgentConfigEngine._load_yaml(rule_path)

            # Ensure rule data contains necessary fields
            if not all(k in rule_data for k in ["name", "prompt"]):
                raise ValueError(f"Rule {rule_file} missing required fields")
            logger.debug(f"Loading rule from {rule_file}, \n{rule_data}\n")

            rules.append(RuleConfig(**rule_data))
        return rules

    @staticmethod
    def load(config_dir: str) -> AgentConfig:
        """
        Load and build AgentConfig

        Args:
            config_dir: Base path of prompt configuration files, e.g. "prompts/CoachLi/v1"

        Returns:
            AgentConfig: Built AgentConfig object

        Raises:
            FileNotFoundError: When necessary configuration files are not found
            ValueError: When configuration file format is incorrect
        """
        if not config_dir or not os.path.exists(config_dir):
            raise ValueError(f"Invalid config directory: {config_dir}")

        # 1. Load meta configuration
        meta_path = os.path.join(config_dir, "meta.yaml")
        if not os.path.exists(meta_path):
            raise FileNotFoundError(f"Meta configuration file not found: {meta_path}")

        meta_data = AgentConfigEngine._load_yaml(meta_path)

        # 2. Build AgentMeta
        agent_meta = AgentMeta(**meta_data.get("meta", {}))

        # 3. Load and render system prompt
        system_prompt = AgentConfigEngine._load_system_prompt(config_dir, meta_data)

        # 4. Load rules
        rules = AgentConfigEngine._load_rules(config_dir)

        # 5. Build and return AgentConfig
        return AgentConfig(
            name=meta_data.get("agent", {}).get("name", "Unnamed Agent"),
            meta=agent_meta,
            system_prompt=system_prompt,
            rules=rules,
            max_concurrent_tasks=meta_data.get("agent", {}).get("max_concurrent_tasks", 1),
            timeout_detection_sec=meta_data.get("agent", {}).get("timeout_detection_sec", 60),
            stream_registration_timeout=meta_data.get("agent", {}).get("stream_registration_timeout", 1),
        )



import enum
from typing import Optional
from dataclasses import dataclass
from dataclasses_json import DataClassJsonMixin
from .rule_config import RuleConfig

class TaskState(enum.Enum):
    """
    Task state enum

    PENDING: Initial state, task is created but not ready to execute
    READY: Conditions are met, task is ready to execute
    EXECUTING: Task is executing
    COMPLETED: Task is completed
    FAILED: Task execution failed
    """
    PENDING = "pending"     # Initial state
    READY = "ready"         # Conditions are met, task is ready to execute
    EXECUTING = "executing" # Task is executing
    COMPLETED = "completed" # Task is completed
    FAILED = "failed"       # Task execution failed

@dataclass
class TaskInfo(DataClassJsonMixin):
    task_id: str
    rule_id: str
    rule_config: RuleConfig
    state: TaskState
    success: bool
    error: Optional[str]
    execution_time: Optional[float]
    response_full: Optional[str]
    created_at: float
    updated_at: float


class TaskStateTransitionError(Exception):
    """
    Task state transition error

    Raises when an invalid state transition is attempted
    """
    def __init__(self, current_state: TaskState, target_state: TaskState, message: Optional[str] = None):
        self.current_state = current_state
        self.target_state = target_state
        msg = f"Invalid state transition: {current_state.value} -> {target_state.value}"
        if message:
            msg += f": {message}"
        super().__init__(msg)



# Copy from https://github.com/Intelligent-Internet/ii-agent

from pathlib import Path
from typing import Optional


class WorkspaceManager:
    root: Path

    def __init__(self, root: Path, container_workspace: Optional[Path] = None):
        self.root = root.absolute()
        self.container_workspace = container_workspace

    def workspace_path(self, path: Path | str) -> Path:
        """Given a path, possibly in a container workspace, return the absolute local path."""
        path = Path(path)
        if not path.is_absolute():
            return self.root / path
        if self.container_workspace and path.is_relative_to(self.container_workspace):
            return self.root / path.relative_to(self.container_workspace)
        return path

    def container_path(self, path: Path | str) -> Path:
        """Given a path, possibly in the local workspace, return the absolute container path.
        If there is no container workspace, return the absolute local path.
        """
        path = Path(path)
        if not path.is_absolute():
            if self.container_workspace:
                return self.container_workspace / path
            else:
                return self.root / path
        if self.container_workspace and path.is_relative_to(self.root):
            return self.container_workspace / path.relative_to(self.root)
        return path

    def relative_path(self, path: Path | str) -> Path:
        """Given a path, return the relative path from the workspace root.
        If the path is not under the workspace root, returns the absolute path.
        """
        path = Path(path)
        abs_path = self.workspace_path(path)
        try:
            return abs_path.relative_to(self.root.absolute())
        except ValueError:
            return abs_path




from typing import Dict, Optional, List
import asyncio
import time
import traceback
import uuid
import loguru
from loguru._logger import Logger

from .context_manager import ContextManager
from .task_executor import TaskExecutor
from .event_bus import EventType, RealTimeEvent
from .output_stream_manager import OutputStreamManager
from .task_state import TaskState, TaskInfo
from .rule_config import RuleConfig



valid_transitions = {
    TaskState.PENDING: [TaskState.READY, TaskState.FAILED],
    TaskState.READY: [TaskState.EXECUTING, TaskState.FAILED],
    TaskState.EXECUTING: [TaskState.COMPLETED, TaskState.FAILED],
    TaskState.COMPLETED: [],  # Terminal state, cannot transition
    TaskState.FAILED: [],  # Terminal state, cannot transition
}


class TaskManager:
    """
    Task manager, responsible for task creation, state management, and lifecycle monitoring
    """

    def __init__(
        self,
        context_manager: ContextManager,
        executor: TaskExecutor,
        output_manager: OutputStreamManager,
        logger: Logger,
        max_concurrent_tasks: int = 10,
        timeout_detection_sec: int = 60,
    ):
        self._context_manager: ContextManager = context_manager
        self._executor: TaskExecutor = executor
        self._output_manager: OutputStreamManager = output_manager
        self._logger: Logger = logger
        self._tasks: Dict[str, TaskInfo] = {}  # task_id -> task_info
        self._task_semaphore = asyncio.Semaphore(max_concurrent_tasks)
        # Task monitor
        self._monitor_task: Optional[asyncio.Task] = None
        # Timeout detection
        self._timeout_detection_sec: int = timeout_detection_sec
        # Start monitor task
        self._start_monitor()

        # Subscribe to task events
        self._context_manager.subscribe(EventType.RULE_TASK_EXECUTING, self._handle_task_executing)
        self._context_manager.subscribe(EventType.RULE_TASK_COMPLETED, self._handle_task_completed)
        self._context_manager.subscribe(EventType.RULE_TASK_FAILED, self._handle_task_failed)
        self._context_manager.subscribe(EventType.RULE_TASK_CANCELLED, self._handle_task_cancelled)

    async def _handle_task_executing(self, event: RealTimeEvent):
        self._logger.info(f"Task {event.data['task_id']} is executing")

    async def _handle_task_completed(self, event: RealTimeEvent):
        self.release_execution_slot()
        self._logger.info(f"Task {event.data['task_id']} is completed")

    async def _handle_task_failed(self, event: RealTimeEvent):
        self.release_execution_slot()
        self._logger.info(f"Task {event.data['task_id']} is failed")

    async def _handle_task_cancelled(self, event: RealTimeEvent):
        self.release_execution_slot()
        self._logger.info(f"Task {event.data['task_id']} is cancelled")


    def generate_task_id(self, rule_id: str) -> str:
        """Generate a unique task ID"""
        return f"task_{rule_id}_{str(uuid.uuid4())[:8]}"

    async def create_task_and_schedule(self, rule_id: str, rule_config: RuleConfig) -> str:
        """Create a new task"""
        task_id = self.generate_task_id(rule_id)
        new_task: TaskInfo = TaskInfo(
            task_id=task_id,
            rule_id=rule_id,
            rule_config=rule_config,
            state=TaskState.PENDING,
            success=False,
            error=None,
            execution_time=None,
            response_full=None,
            created_at=time.time(),
            updated_at=time.time(),
        )
        self._tasks[task_id] = new_task
        # Schedule task
        await self._schedule_task(task_id)

        return task_id

    async def _schedule_task(self, task_id: str) -> None:
        """Schedule task execution"""
        # Try to acquire an execution slot
        if not await self.acquire_execution_slot():
            self._logger.info(f"Max concurrent tasks reached. Delaying task {task_id}")
            return
        if task_id not in self._tasks:
            self._logger.error(f"Task {task_id} not found, skip scheduling")
            return
        task_info = self._tasks[task_id]
        if task_info.state != TaskState.PENDING:
            self._logger.error(f"Task {task_id} is not in PENDING state, skip scheduling")
            return
        try:
            # Set task state to ready
            await self._set_task_state(task_id, TaskState.READY)
            task_info: TaskInfo = self.get_task_info(task_id)
            rule_config = task_info.rule_config
            # Get async generator
            async_generator = self._executor.run_task(task_info)
            # Register this stream with output manager
            stream_id = self._output_manager.register_stream(
                async_generator, task_info
            )

            self._logger.info(
                f"Task {task_id} (Name: {rule_config.name}) is scheduled for execution. stream_id={stream_id}"
            )
        except Exception as e:
            self._logger.error(f"Error scheduling task {task_id}: {e} {traceback.format_exc()}")
            self.release_execution_slot()
            await self._set_task_state(task_id, TaskState.FAILED, success=False, error=str(e))
            

        
    async def _set_task_state(
        self,
        task_id: str,
        new_state: TaskState,
        success: Optional[bool] = None,
        error: Optional[str] = None,
    ) -> None:
        """Set task state"""
        if task_id not in self._tasks:
            self._logger.error(f"Task {task_id} not found")
            return

        task_info = self._tasks[task_id]
        old_state: TaskState = task_info.state

        # Check if the state transition is valid
        if not self._is_valid_state_transition(old_state, new_state):
            self._logger.error(
                f"Invalid state transition: {old_state} -> {new_state}"
            )
            return

        # Update state
        task_info.state = new_state
        if success is not None:
            task_info.success = success
        if error is not None:
            task_info.error = error
        task_info.updated_at = time.time()
        self._logger.info(
            f"Task {task_id} state changed: {old_state.value} -> {new_state.value}"
        )

        
            
    def _is_valid_state_transition(
        self, current_state: TaskState, new_state: TaskState
    ) -> bool:
        """Check if the state transition is valid"""
        return new_state in valid_transitions[current_state]

    def get_task_state(self, task_id: str) -> Optional[TaskState]:
        """Get task state"""
        if task_id not in self._tasks:
            return None
        return self._tasks[task_id].state

    def is_task_completed(self, task_id: str) -> bool:
        """Check if the task is completed"""
        state = self.get_task_state(task_id)
        return state in (TaskState.COMPLETED, TaskState.FAILED)

    def is_task_executing(self, task_id: str) -> bool:
        """Check if the task is executing"""
        return self.get_task_state(task_id) == TaskState.EXECUTING

    def is_task_ready(self, task_id: str) -> bool:
        """Check if the task is ready"""
        return self.get_task_state(task_id) == TaskState.READY

    def get_task_info(self, task_id: str) -> Optional[TaskInfo]:
        """Get task info"""
        return self._tasks.get(task_id)

    def get_tasks_by_rule(self, rule_id: str) -> List[str]:
        """Get all task IDs for a specific rule"""
        return [
            task_id
            for task_id, task_info in self._tasks.items()
            if task_info.rule_id == rule_id
        ]

    def get_active_tasks(self) -> List[str]:
        """Get all active task IDs"""
        return [
            task_id
            for task_id, task_info in self._tasks.items()
            if task_info.state not in (TaskState.COMPLETED, TaskState.FAILED)
        ]

    async def acquire_execution_slot(self) -> bool:
        """Acquire an execution slot"""
        try:
            # Set timeout to avoid infinite wait
            await asyncio.wait_for(self._task_semaphore.acquire(), timeout=2)
            return True
        except asyncio.TimeoutError:
            return False

    def release_execution_slot(self) -> None:
        """Release an execution slot"""
        self._task_semaphore.release()

    def get_inactive_tasks(self) -> List[str]:
        """Get inactive tasks"""
        current_time = time.time()
        return [
            task_id
            for task_id, task_info in self._tasks.items()
            if not self.is_task_completed(task_id)
            and current_time - task_info.updated_at > self._timeout_detection_sec
        ]

    def _start_monitor(self) -> None:
        """
        Start monitor task
        """
        if self._monitor_task is None or self._monitor_task.done():
            self._monitor_task = asyncio.create_task(self._monitor_loop())

    async def _monitor_loop(self) -> None:
        """
        Monitor loop, check task status and potential timeouts
        """
        try:
            while True:
                await asyncio.sleep(1)  # Check every second

                # Check for inactive tasks
                active_tasks = self.get_active_tasks()
                for task_id in active_tasks:
                    if self.get_task_state(task_id) == TaskState.PENDING:
                        await self._schedule_task(task_id)
                        continue
                    
                    if time.time() - self._tasks[task_id].updated_at > self._timeout_detection_sec:
                        self._logger.warning(
                            f"Task {task_id} has been inactive for {self._timeout_detection_sec} seconds."
                        )
                        # Try to mark the stuck task as failed
                        if self.is_task_executing(task_id):
                            try:
                                await self._executor.cancel_task(task_id)
                                self._logger.warning(
                                        f"Marked stuck task {task_id} as FAILED, error: Timeout detected"
                                )
                            except Exception as e:
                                self._logger.error(
                                    f"Error handling stuck task {task_id}: {e} {traceback.format_exc()}"
                                )
                        
        except asyncio.CancelledError:
            self._logger.debug("Monitor task cancelled")
        except Exception as e:
            self._logger.error(f"Error in monitor loop: {e} {traceback.format_exc()}")

    async def shutdown(self) -> None:
        """Shutdown task manager"""
        # Mark all active tasks as failed
        for task_id in self.get_active_tasks():
            await self._executor.cancel_task(task_id)

        # Cancel monitor task
        if self._monitor_task:
            self._monitor_task.cancel()
            try:
                await self._monitor_task
            except asyncio.CancelledError:
                pass



import loguru
from loguru._logger import Logger

from ..libs.crawler.crawl4ai import Crawl4AIWrapper
from .base import Tool


class Crawl4AITool(Tool):
    name: str = "Crawl4AITool"
    description: str = (
        "A tool can crawl a website url then return a markdown text. "
        "Useful for when you need to answer questions about current events. "
        "Input should be a url. Output is a markdown text."
    )
    
    def __init__(self):
        super().__init__()
        self.wrapper = Crawl4AIWrapper()


    async def run(self, url: str, **kwargs) -> str:
        """
        Run crawl for a url and return markdown text.

        Args:
            url: The url to crawl.

        Returns:
            A markdown text.
        """
        return await self.wrapper.crawl(url=url)


import sys

try:
    import re
    import requests
    from markdownify import markdownify
    from requests.exceptions import RequestException

except ImportError:
    raise Exception("Required packages 'markdownify' and 'requests' are not installed")

MAX_LENGTH_TRUNCATE_CONTENT = 1000

def truncate_content(
    content: str, max_length: int = MAX_LENGTH_TRUNCATE_CONTENT
) -> str:
    if len(content) <= max_length:
        return content
    else:
        return (
            content[: max_length // 2]
            + f"\n..._This content has been truncated to stay below {max_length} characters_...\n"
            + content[-max_length // 2 :]
        )


def main(url: str) -> str:
    try:
        # Send a GET request to the URL with a 20-second timeout
        response = requests.get(url, timeout=20, headers={
            'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7',
            'accept-language': 'en,zh-CN;q=0.9,zh;q=0.8',
            'cache-control': 'max-age=0',
            'oai-did': 'a081b658-bad7-41f2-8ed0-28cf7975a6d7; ajs_anonymous_id=15d3da39-f27b-4ca6-ba02-7d606a81b9b5; intercom-id-dgkjq2bp=6236a5df-3fa1-45e6-bedd-b229178bb43b; intercom-device-id-dgkjq2bp=b24cd420-c14d-4d5d-98ff-0b6514d11609; _ga=GA1.1.471785481.1732446093; __stripe_mid=fffe2718-997f-4404-9b7a-807a2b9bf0929ef9b4; __stripe_mid=fffe2718-997f-4404-9b7a-807a2b9bf0929ef9b4; _ga_8MYC5SEFJ1=GS1.1.1735784365.10.0.1735784365.0.0.0; _ga_62J2E5SERF=GS2.1.s1747732581$o5$g1$t1747735176$j0$l0$h0; __cf_bm=tC6LR3_9rbxUm8Awd.evIh9bK_76Y31KMtyb1dkc00A-1748503882-1.0.1.1-KwH5HwHuDsvZvf9Qz1Lg_3jSY6h_xQyRX2tWyZa7.ymQfnEAReehybcyeYyTsAjWlNI3wwGgry4jSQnbaGtR2frukimWBWq..jCL4nxS7FA; _cfuvid=I58jukf4ZRxBd7sDA1LBF2O3KpeB5Kizt4yEvVJnmn8-1748503882122-0.0.1.1-604800000; cf_clearance=CJI73CKZOxWlrNA1XpPYSkkUohRR9.dtfYDta8Q9cic-1748503884-1.2.1.1-SYKOb671HEH8fNFi31kN6OBOC0p4MSdiwtCwVkoME.P7flfz7EKdXpzVkGr7vmhUmqDbWR9I8F4mgTHpJGx4UxnpdxDltyLSyvBh8_CDLbxW_G177YerqpArLtjC7NyZmLakQgSgxK8OaQa3ne6TIwz6oRhJ58MrGKT.ryew1CcMXvCZLFe8xEe31zND.HMpFL60Kl5EMf4cRIou490.sq63xnYriOJ8XxObGvMgy3cwxtXZK3FJ5GRl.wOcsqm1eVkNk642XJXK1RDsuxDWO7cGDecNeSfyY1RSxq7_KHvb9prngmkRu3J7FV8yTxUUMPIM0HFyGqtCDDplMyVGXjyiYCk7wZfbG2tjSpo.MeM; intercom-session-dgkjq2bp=; _dd_s=logs=1&id=0c098b66-0873-48c7-a256-ebf59742bb4d&created=1748503883294&expire=1748505083301',
            'priority': 'u=0, i',
            'sec-ch-ua': '"Chromium";v="136", "Google Chrome";v="136", "Not.A/Brand";v="99"',
            'sec-ch-ua-mobile': '0',
            'sec-ch-ua-platform': 'macOS',
            'sec-fetch-dest': 'document',
            'sec-fetch-mode': 'navigate',
            'sec-fetch-site': 'none',
            'sec-fetch-user': '?1',
            'upgrade-insecure-requests': '1',
            'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/136.0.0.0 Safari/537.3'
        })
        response.raise_for_status()

        # Convert the HTML content to Markdown
        print(response.text)
        markdown_content = markdownify(response.text).strip()

        # Remove multiple line breaks
        markdown_content = re.sub(r"\n{3,}", "\n\n", markdown_content)

        if not markdown_content:
            raise Exception("No content found in the webpage")

        return markdown_content

    except requests.exceptions.Timeout:
        raise Exception("The request timed out")
    except RequestException as e:
        raise Exception(f"Error fetching the webpage: {str(e)}")


if __name__ == "__main__":
    args = sys.argv
    if len(args) != 2:
        print("Usage: python test_markdownify.py <url>")
        sys.exit(1)
    url = args[1]
    print(main(url))



import asyncio
import loguru
from loguru._logger import Logger
from crawl4ai import *
from crawl4ai import content_filter_strategy
from crawl4ai.models import RunManyReturn


class Crawl4AIWrapper(BaseModel):
    """Wrapper for Crawl4AI"""

    browser_config: BrowserConfig
    run_config: CrawlerRunConfig

    def __init__(self):
        self.browser_config = BrowserConfig(
            verbose=False,
            headless=True,
            light_mode=True,
            # proxy="http://127.0.0.1:7890",
        )  # Default browser configuration

        self.content_filter = PruningContentFilter(threshold=0.4)
        # Create a markdown generator with a content filter
        self.md_generator = DefaultMarkdownGenerator(
            content_filter=self.content_filter,
            options={
                "body_width": 88,  # Wrap text at 80 characters
                "ignore_images": True,  # Skip images
                "single_line_break": True,  # Use single line breaks
            },
        )
        self.run_config = CrawlerRunConfig(
            # Content filtering
            word_count_threshold=10,
            excluded_tags=["form", "header"],
            exclude_external_links=True,
            # Content processing
            process_iframes=True,
            remove_overlay_elements=True,
            # Cache control
            cache_mode=CacheMode.ENABLED,  # Use cache if available)   # Default crawl run configuration
            markdown_generator=self.md_generator,
        )

    async def crawl(self, url: str):
        async with AsyncWebCrawler(config=self.browser_config) as crawler:
            result = await crawler.arun(
                url=url,
                config=self.run_config,
            )
            # Check success status
            if not result.success:
                loguru.logger.error(result.error)
                return ""

            # print(result.cleaned_html)  # Cleaned HTML
            if result.markdown:
                loguru.logger.info("--==--" * 50 + "\n RAW MARKDOWN\n")
                loguru.logger.info(
                    result.markdown.raw_markdown
                )  # Raw markdown from cleaned html
                loguru.logger.info("--==--" * 50 + "\n FIT MARKDOWN\n")
                loguru.logger.info(
                    result.markdown.fit_markdown
                )  # Most relevant content in markdown
                return (
                    result.markdown.fit_markdown
                    if result.markdown.fit_markdown
                    else result.markdown.raw_markdown
                )

            else:
                return result.cleaned_html


if __name__ == "__main__":
    wrapper = Crawl4AIWrapper()
    asyncio.run(
        wrapper.crawl(
            url="https://deepwiki.com/unclecode/crawl4ai/7.2-configuration-best-practices"
        )
    )



import time
import loguru
from loguru._logger import Logger
from typing import Callable, Optional
import asyncio
from .event_bus import EventBus, RealTimeEvent, EventType
from .observable_ctx import ObservableCtx
from .llm.llm_base import (
    GeneralContentBlock,
    TextPrompt,
    ToolCall,
    ToolResult,
    TextResult,
    TokenUsage,
)
from .task_state import TaskState, TaskInfo
from .rule_config import RuleConfig


class ContextManager:

    def __init__(self, logger: Logger):
        self._event_bus = EventBus(logger=logger) 
        self._ctx = ObservableCtx(self._event_bus, logger) 
        self._message_history = asyncio.PriorityQueue()
        self._logger = logger

    def get_context(self) -> ObservableCtx:
        return self._ctx

    def get_event_bus(self) -> EventBus:
        return self._event_bus

    def subscribe(
        self, event_type: EventType, callback: Callable[[RealTimeEvent], None]
    ):
        self._logger.info(f"Subscribe to event {event_type}, callback {callback.__name__}")
        self._event_bus.subscribe(event_type, callback)

    def set_system_prompt(self, system_prompt: str):
        self._ctx.set("system_prompt", system_prompt)

    def emit_event(self, event: RealTimeEvent):
        self._logger.info(f"Emit event {event.type}, data {event.model_dump()}")
        self._event_bus.emit(event)

    async def emit_and_append_to_history(self, message: GeneralContentBlock):
        """
        Send event and add to message history
        """
        await self._message_history.put((time.time(), message))
        self._ctx.append("dialogue.history", message.to_dict())
        if isinstance(message, TextPrompt):
            self._ctx.set("user_input", message.text)
            if message.image_url:
                self._ctx.set("image_url", message.image_url)
            self.emit_event(
                RealTimeEvent(type=EventType.USER_MESSAGE, data=message.to_dict())
            )
        elif isinstance(message, ToolCall):
            self.emit_event(
                RealTimeEvent(type=EventType.TOOL_CALL, data=message.to_dict())
            )
        elif isinstance(message, ToolResult):
            self.emit_event(
                RealTimeEvent(type=EventType.TOOL_RESULT, data=message.to_dict())
            )
        elif isinstance(message, TextResult):
            self.emit_event(
                RealTimeEvent(type=EventType.AGENT_RESPONSE, data=message.to_dict())
            )
        elif isinstance(message, TokenUsage):
            self.emit_event(
                RealTimeEvent(type=EventType.TOKEN_USAGE, data=message.to_dict())
            )

    def emit_task_executing(self, task_info: TaskInfo):
        task_info.state = TaskState.EXECUTING
        task_info.updated_at = time.time()
        self.emit_event(
            RealTimeEvent(type=EventType.RULE_TASK_EXECUTING, data=task_info.to_dict())
        )

    def emit_task_cancelled(self, task_info: TaskInfo, error_msg: Optional[str]):
        task_info.state = TaskState.FAILED
        task_info.success = False
        task_info.error = error_msg
        task_info.updated_at = time.time()
        self.emit_event(
            RealTimeEvent(
                type=EventType.RULE_TASK_CANCELLED,
                data=task_info.to_dict(),
            )
        )

    def emit_task_failed(
        self, task_info: TaskInfo, execution_time: float, error_msg: Optional[str]
    ):
        task_info.state = TaskState.FAILED
        task_info.success = False
        task_info.error = error_msg
        task_info.execution_time = execution_time
        task_info.updated_at = time.time()
        self.emit_event(
            RealTimeEvent(
                type=EventType.RULE_TASK_FAILED,
                data=task_info.to_dict(),
            )
        )

    def emit_task_completed(
        self, task_info: TaskInfo, execution_time: float, response_full: Optional[str]
    ):
        task_info.state = TaskState.COMPLETED
        task_info.success = True
        task_info.error = None
        task_info.execution_time = execution_time
        task_info.response_full = response_full
        task_info.updated_at = time.time()
        self.emit_event(
            RealTimeEvent(
                type=EventType.RULE_TASK_COMPLETED,
                data=task_info.to_dict(),
            )
        )

    def emit_task_generate_new_rule(
        self, task_info: TaskInfo, rule_config: RuleConfig, immediate: bool = False
    ):
        self.emit_event(
            RealTimeEvent(
                type=EventType.NEW_RULE_GENERATED,
                data={
                    "task_id": task_info.task_id,
                    "rule_config": rule_config,
                    "immediate": immediate,
                },
            )
        )



"""
Tool base class and registry
"""

import functools
import inspect
import loguru
from loguru._logger import Logger
from abc import ABC, abstractmethod
from typing import Any, AsyncGenerator, Callable, Dict, List, Optional, Type, Union
from docstring_parser import parse
from .utils import get_json_type_info


class Tool(ABC):
    """
    Tool abstract base class
    """

    name: str = ""
    description: str = ""

    @abstractmethod
    async def run(self, *args, **kwargs) -> AsyncGenerator[str, None]:
        """
        The base interface for tools, all tools should implement this interface
        Return an async generator to stream the result
        """
        raise NotImplementedError


class ToolCenter:
    """
    Tool registry, used to manage and access available tools
    """

    _tools: Dict[str, Union[Tool, Callable]] = {}

    @staticmethod
    def register(
        tool: Union[Tool, Callable, Type[Tool], None] = None,
        *,
        name: Optional[str] = None,
    ) -> Union[
        Callable[
            [Union[Tool, Callable, Type[Tool]]], Union[Tool, Callable, Type[Tool]]
        ],
        None,
    ]:
        """
        Register a tool, supports two ways of use:
        1. As a regular method: ToolCenter.register(tool, name="tool_name")
        2. As a decorator: @ToolCenter.register or @ToolCenter.register(name="tool_name")

        Args:
            tool: Tool instance, function or tool class
            name: Tool name, if None, use tool's name attribute or function name

        Returns:
            When used as a decorator, return the decorated function or class; otherwise return None
        """
        # As a decorator usage, no parameter form: @ToolCenter.register
        if tool is not None and name is None:
            return ToolCenter._register_tool(tool, name)

        # As a decorator usage, with parameter form: @ToolCenter.register(name="tool_name")
        if tool is None:

            def decorator(
                tool_func: Union[Tool, Callable, Type[Tool]]
            ) -> Union[Tool, Callable, Type[Tool]]:
                ToolCenter._register_tool(tool_func, name)
                return tool_func

            return decorator

        # As a regular method usage: ToolCenter.register(tool, name="tool_name")
        return ToolCenter._register_tool(tool, name)

    @staticmethod
    def _register_tool(
        tool: Union[Tool, Callable, Type[Tool]], name: Optional[str] = None
    ) -> None:
        """
        Internal method for actual tool registration
        """
        if isinstance(tool, type) and issubclass(tool, Tool):
            # If it's a tool class, instantiate it
            tool_instance = tool()
            tool_name = name or tool_instance.name or tool.__name__
            ToolCenter._tools[tool_name] = tool_instance
            loguru.logger.info(f"Register tool class: {tool_name}")

        elif isinstance(tool, Tool):
            # If it's a tool instance
            tool_name = name or tool.name or tool.__class__.__name__
            ToolCenter._tools[tool_name] = tool
            loguru.logger.info(f"Register tool instance: {tool_name}")

        elif callable(tool):
            # If it's a function
            tool_name = name or tool.__name__
            ToolCenter._tools[tool_name] = tool
            loguru.logger.info(f"Register tool function: {tool_name}")

        else:
            raise TypeError(f"Unsupported tool type: {type(tool)}")

        return None

    @staticmethod
    def get(name: str) -> Optional[Union[Tool, Callable]]:
        """
        Get a tool by name

        Args:
            name: Tool name

        Returns:
            Tool instance or function, returns None if not found
        """
        return ToolCenter._tools.get(name)

    @staticmethod
    @functools.lru_cache(maxsize=1000)
    def list_tools() -> List[str]:
        """
        List all available tool names

        Returns:
            List of tool names
        """
        return list(ToolCenter._tools.keys())

    @staticmethod
    @functools.lru_cache(maxsize=1000)
    def get_definition(name: str) -> Optional[Dict[str, Any]]:
        """
        Dynamically build OpenAI function call tool definition (optimized version)
        """
        tool = ToolCenter.get(name)
        if tool is None:
            return None

        # 1. Get callable object (prefer run method, otherwise use tool object directly)
        func: Callable = tool.run if hasattr(tool, "run") else tool
        if not callable(func):
            return None

        # 2. Unwrap decorators (if function is wrapped by decorators, get the original function)
        try:
            original_func = inspect.unwrap(func)  # Unwrap decorators
            sig = inspect.signature(original_func)
        except ValueError:
            sig = inspect.signature(func)  # If cannot unwrap, use current function signature

        # 3. Parse docstring parameters (support multiple styles)
        docstring = inspect.getdoc(func) or ""
        parsed_doc = parse(docstring)  # Parsed docstring object
        param_docs = {
            p.arg_name: p.description for p in parsed_doc.params
        }  # Parameter name to description mapping

        # 4. Build parameters structure
        properties: Dict[str, Dict[str, Any]] = {}
        required: list[str] = []

        for pname, param in sig.parameters.items():
            # Skip special parameters (self/cls/args/kwargs)
            if pname in ("self", "cls", "args", "kwargs"):
                continue

            # -------------------- Type inference --------------------
            annotation = param.annotation
            json_type_info = get_json_type_info(
                annotation
            )  # Get JSON schema type information (supports generics)

            # -------------------- Parameter description --------------------
            # Prefer docstring parameter description, then type annotation __doc__, finally empty
            desc = param_docs.get(pname, "")
            if not desc and hasattr(annotation, "__doc__"):
                desc = annotation.__doc__ or ""

            # -------------------- Assemble properties --------------------
            properties[pname] = {
                **json_type_info,  # Contains "type" or "type"+"items" (e.g., array type)
                "description": desc.strip(),  # Remove leading and trailing whitespace
            }

            # -------------------- Determine if required --------------------
            if param.default is inspect.Parameter.empty:
                required.append(pname)

        # 5. Build final schema (OpenAI function call specification)
        schema = {
            "name": name,
            "description": parsed_doc.short_description
            or name,  # Use docstring short description
            "parameters": {
                "type": "object",
                "properties": properties,
                "required": required if required else None,  # Set to None if no required parameters (optional)
            },
        }

        # Remove possible None in parameters (OpenAI requires fields to be present)
        if schema["parameters"]["required"] is None:
            del schema["parameters"]["required"]

        return schema

    @staticmethod
    async def run_tool(name: str, *args, **kwargs) -> AsyncGenerator[str, None]:
        """
        Run a specified tool

        Args:
            name: Tool name
            *args: Positional arguments
            **kwargs: Keyword arguments

        Yields:
            Result stream from the tool execution

        Raises:
            ValueError: When the tool does not exist
            TypeError: When the tool type is not supported
        """
        tool = ToolCenter.get(name)

        if tool is None:
            raise ValueError(f"Tool not found: {name}")

        if isinstance(tool, Tool):
            # If it's a tool instance, call its run method
            async for chunk in tool.run(*args, **kwargs):
                yield chunk

        elif inspect.isasyncgenfunction(tool):
            # If it's an async generator function
            async for chunk in tool(*args, **kwargs):
                yield chunk

        elif inspect.iscoroutinefunction(tool):
            # If it's an async function, return its result as a single chunk
            result = await tool(*args, **kwargs)
            yield str(result)

        else:
            raise TypeError(f"Unsupported tool type: {type(tool)}")



from .base import Tool


class CodeReplTool(Tool):
    """
    Code Repl Tool, used to execute code and return result
    """

    def __init__(self, **kwargs):
        pass

    def run(self, **kwargs):
        pass



# Copy from https://github.com/Intelligent-Internet/ii-agent

from dataclasses import dataclass
from typing import Any, Optional
from dataclasses_json import DataClassJsonMixin
from typing import Literal


@dataclass
class ToolCallParameters:
    tool_call_id: str
    tool_name: str
    tool_input: Any


@dataclass
class ToolParam(DataClassJsonMixin):
    """Internal representation of LLM tool."""

    name: str
    description: str
    input_schema: dict[str, Any]


@dataclass
class ToolCall(DataClassJsonMixin):
    """Internal representation of LLM-generated tool call."""

    tool_call_id: str
    tool_name: str
    tool_input: Any


@dataclass
class ToolResult(DataClassJsonMixin):
    """Internal representation of LLM tool result."""

    tool_call_id: str
    tool_name: str
    tool_output: Any


@dataclass
class ToolFormattedResult(DataClassJsonMixin):
    """Internal representation of formatted LLM tool result."""

    tool_call_id: str
    tool_name: str
    tool_output: list[dict[str, Any]] | str


@dataclass
class TextPrompt(DataClassJsonMixin):
    """Internal representation of user-generated text prompt."""

    text: str
    image_url: Optional[str] = None
    image_url_detail: Optional[str] = "high"
    role: Literal["user", "assistant"] = "user"


@dataclass
class ImageBlock(DataClassJsonMixin):
    type: Literal["image"]
    source: dict[str, Any]
    role: Literal["assistant"] = "assistant"


@dataclass
class AIContext(DataClassJsonMixin):
    """Internal representation of LLM-generated middle result."""

    key: str
    context: str
    role: Literal["assistant"] = "assistant"


@dataclass
class TextResult(DataClassJsonMixin):
    """Internal representation of LLM-generated text result."""

    text: str
    role: Literal["assistant"] = "assistant"


@dataclass
class TokenUsage(DataClassJsonMixin):
    """Internal representation of LLM token usage."""

    model: str
    total_tokens: int
    prompt_tokens: int
    completion_tokens: int
    completion_tokens_details: Optional[dict[str, Any]]
    prompt_tokens_details: Optional[dict[str, Any]]
    role: Literal["user", "assistant"] = "assistant"


# agent content block
AssistantContentBlock = TextResult | ToolCall | AIContext | TokenUsage

# user content block
UserContentBlock = TextPrompt | ToolFormattedResult

# general content block
GeneralContentBlock = UserContentBlock | AssistantContentBlock

# LLM messages
LLMMessages = list[list[GeneralContentBlock]]



import asyncio
from crawl4ai import *
from crawl4ai import content_filter_strategy
from crawl4ai.models import RunManyReturn
# from crawl4ai.markdown import DefaultMarkdownGenerator, BM25ContentFilter, PruningContentFilter


async def main():

    browser_config = BrowserConfig(
        verbose=True,
        headless=True,
        light_mode=True,
        proxy="http://127.0.0.1:7890",
    )  # Default browser configuration
    
    # content_filter = BM25ContentFilter(
    #     user_query="Anthropic",  # Optional query to focus the content
    #     bm25_threshold=1.0,                    # Minimum score threshold
    #     language="english"                     # Language for stemming
    # )
    content_filter = PruningContentFilter(threshold=0.4)
    
    # Create a markdown generator with a content filter
    md_generator = DefaultMarkdownGenerator(
        content_filter=content_filter,
        options={
            "body_width": 80,  # Wrap text at 80 characters
            "ignore_images": True,  # Skip images
            "single_line_break": True,  # Use single line breaks
        }
    )

    run_config = CrawlerRunConfig(
        # Content filtering
        word_count_threshold=10,
        excluded_tags=["form", "header"],
        exclude_external_links=True,
        # Content processing
        process_iframes=True,
        remove_overlay_elements=True,
        # Cache control
        cache_mode=CacheMode.ENABLED,  # Use cache if available)   # Default crawl run configuration
        markdown_generator=md_generator,
    )

    async with AsyncWebCrawler(config=browser_config) as crawler:
        result = await crawler.arun(
            url="https://deepwiki.com/unclecode/crawl4ai/7.2-configuration-best-practices",
            config=run_config,
        )
        # Check success status
        print(result.success)  # True if crawl succeeded
        if not result.success:
            print(result.error)
            return
        # Different content formats
        # print(result.html)  # Raw HTML
        print("--==--" * 50)
        # print(result.cleaned_html)  # Cleaned HTML
        if result.markdown:
            print("--==--" * 50 + "\n RAW MARKDOWN\n")
            print(result.markdown.raw_markdown)  # Raw markdown from cleaned html
            print("--==--" * 50 + "\n FIT MARKDOWN\n")
            print(result.markdown.fit_markdown)  # Most relevant content in markdown

        


if __name__ == "__main__":
    asyncio.run(main())



from typing import Any, Dict, Optional
import loguru
from loguru._logger import Logger
import json

from .utils import get_nested, set_nested
from .event_bus import EventBus, EventType, RealTimeEvent


class Context:
    def __init__(self):
        self._storage: Dict[str, Any] = {}

    def set(self, key: str, value: Any):
        try:
            value = json.loads(value, strict=False)
        except:
            pass
        set_nested(self._storage, key, value)

    def append(self, key: str, value: Any):
        old_value = self.get(key)
        if old_value is None:
            old_value = []
        if not isinstance(old_value, list):
            raise ValueError(
                f"Key '{key}' is not a list. Cannot append to non-list value."
            )
        old_value.append(value)
        self.set(key, old_value)
        logger.debug(f"Ctx append: key='{key}', value='{value}'")

    def get(self, key: str) -> Any:
        return get_nested(self._storage, key)

    def has(self, key: str) -> bool:
        return get_nested(self._storage, key) is not None

    def to_dict(self) -> Dict[str, Any]:
        return dict(self._storage)


class ObservableCtx:

    def __init__(self, event_bus: EventBus, logger: Logger):
        self._event_bus: EventBus = event_bus
        self._logger: Logger = logger
        self._data: Context = Context()

    def set(self, key: str, value: Any) -> None:
        old_value = self._data.get(key)
        self._data.set(key, value)
        self._logger.info(
            f"Ctx set: key='{key}', value='{value}', old_value='{old_value}'"
        )
        self._event_bus.emit(
            RealTimeEvent(
                type=EventType.CONTEXT_CHANGED,
                data={"key": key, "value": value, "old_value": old_value},
            )
        )

    def append(self, key: str, value: Any) -> None:
        old_value = self._data.get(key)
        if old_value is None:
            old_value = []
        if not isinstance(old_value, list):
            raise ValueError(
                f"Key '{key}' is not a list. Cannot append to non-list value."
            )
        new_value = old_value.copy()
        new_value.append(value)
        self._data.set(key, new_value)
        self._logger.debug(
            f"Ctx append: key='{key}', value='{new_value}', old_value='{old_value}'"
        )
        self._event_bus.emit(
            RealTimeEvent(
                type=EventType.CONTEXT_CHANGED,
                data={"key": key, "value": new_value, "old_value": old_value},
            )
        )

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        value = self._data.get(key)
        if value is None:
            return default
        return value

    def to_dict(self) -> Dict[str, Any]:
        return self._data.to_dict()

    def __contains__(self, key: str) -> bool:
        return self._data.has(key=key)

    def __getitem__(self, key: str) -> Any:
        return self._data.get(key)

    def __setitem__(self, key: str, value: Any) -> None:
        # This doesn't publish event, use set() for that
        # For direct dict-like assignment if needed, but prefer set()
        # To make it publish, call self.set() here, but be mindful of async context
        # For simplicity, direct assignment won't publish. Use `await ctx.set()`
        self._logger.warning(
            f"Direct assignment to ObservableCtx key '{key}'. Event not published. Use await ctx.set()."
        )
        self._data.set(key, value)



from typing import get_origin, get_args
from inspect import Parameter
from typing import Dict, Any


def get_json_type_info(annotation) -> Dict[str, Any]:
    """
    Map Python type annotations to JSON schema type information (supports generics)
    Return examples:
    - Simple types: {"type": "integer"}
    - Generic list: {"type": "array", "items": {"type": "string"}}
    """
    # Default to string when no type annotation (can be adjusted based on needs)
    if annotation is Parameter.empty:
        return {"type": "string"}

    # Handle generic types (e.g., list[str], dict[str, int])
    origin = get_origin(annotation)
    args = get_args(annotation)

    # -------------------- Handle list generic --------------------
    if origin is list:
        if args:  # list[str] → {"type": "array", "items": {"type": "string"}}
            items_type = get_json_type_info(args[0])
            return {"type": "array", "items": items_type}
        else:  # list → {"type": "array"}（No element type information）
            return {"type": "array"}

    # -------------------- Handle dict generic --------------------
    if origin is dict:
        if (
            len(args) >= 2
        ):  # dict[str, int] → {"type": "object", "properties": ...}（Simplified processing）
            return {
                "type": "object"
            }
        else:
            return {"type": "object"}

    # -------------------- Handle built-in types --------------------
    type_mapping = {
        int: "integer",
        float: "number",
        bool: "boolean",
        str: "string",
        dict: "object",
        list: "array",
        tuple: "array",  # Simplified to array
    }
    for py_type, json_type in type_mapping.items():
        if annotation is py_type:
            return {"type": json_type}

    # -------------------- Handle custom classes --------------------
    # Custom classes are treated as object by default (can be extended based on needs, e.g., check if Pydantic model)
    return {"type": "object"}



"""
LLM client base class
"""

from abc import ABC, abstractmethod
from typing import AsyncGenerator


class LLMClient(ABC):
    """
    LLM client abstract base class
    """
    
    
    @abstractmethod
    async def invoke_stream(self, system_prompt: str, user_input: str, **kwargs) -> AsyncGenerator[str, None]:
        """
        Streamly invoke LLM and return response stream
        
        Args:
            system_prompt: System prompt
            user_input: User input
            **kwargs: Other parameters
            
        Returns:
            Async generator, yield response fragments
        """
        pass



from .base import Tool


class RAGTool(Tool):
    """
    RAG retrieval tool, used to retrieve relevant documents from vector database
    """

    def __init__(self, **kwargs):
        pass

    def run(self, **kwargs):
        pass



import json
import os
import time
import loguru
from loguru._logger import Logger

from typing import Optional, Literal, List, Dict, Any
from pydantic import BaseModel

logger = loguru.logger


class AgentMeta(BaseModel):
    name: str
    desc: str
    parameters: Optional[Dict[str, Any]] = None


def save_rule_config(source: str, rule_config: "RuleConfig"):
    """
    Save rule config to file
    """
    # filename: <source>_generated_<name>_<timestamp>.json
    # content: rule_config.model_dump_json()
    # path: logs/rule_configs/
    # create directory if not exists
    os.makedirs("logs/rule_configs", exist_ok=True)
    with open(
        f"logs/rule_configs/{source}_generated_{rule_config.name}_{time.strftime('%Y%m%d_%H%M%S')}.json",
        "w",
    ) as f:
        f.write(rule_config.model_dump_json())


def find_json_in_str(s: str) -> str:
    """
    Find the first JSON object in the string
    """
    start = s.find("{")
    end = s.rfind("}")
    if start != -1 and end != -1:
        return s[start : end + 1]
    return None


class RuleConfig(BaseModel):
    """
    Rule configuration

    Parameters:
        name: Rule name
        depend_ctx_key: The key of the data to be retrieved from the context
        match_condition: Condition to be satisfied, will be executed by eval,
                        for example: "力量训练" in ctx.get("intent_analysis_result.intent")

        prompt: Custom prompt for this rule (will be appended to system_prompt)
        prompt_detail: Additional prompt information for this rule, will be appended to system_prompt + prompt

        tool: The name of the tool to be executed
        tool_params: Parameters of the tools to be used
                    ```json
                    {
                        "ctx": ["user_input"], # dynamically generated by other Rule or Agent
                        "config": ["prompt", "prompt_detail"], # fixed configuration
                        "extra": {}, # additional constants
                    }
                    ```
        tool_result_target: Where the AI output of this rule should be sent to:
                            - AS_CONTEXT store in context
                            - DIRECT_RETURN return to user directly
                            - NEW_RULES generate new rules
        tool_result_key: If AS_CONTEXT, the key to store the AI output in ctx
        auto_generated: Whether this rule is automatically generated
        priority: Priority
    """

    name: str
    depend_ctx_key: List[str]
    match_condition: Optional[str] = None

    prompt: str
    prompt_detail: Optional[str] = ""

    tool: Optional[str] = None
    tool_params: Optional[Dict[str, Any]] = {}
    tool_result_target: Literal["AS_CONTEXT", "DIRECT_RETURN", "NEW_RULES"]
    tool_result_key: Optional[str] = None
    auto_generated: bool = False
    priority: int = 0

    @classmethod
    def parse_and_gen(
        cls, source: str, tool_result_full: str, save: bool = False
    ) -> List["RuleConfig"]:
        """
        Parse llm_response_full and generate new rules
        """
        try:
            # Try to parse the full JSON
            json_data = json.loads(tool_result_full, strict=False)
        except json.JSONDecodeError:
            # Try to extract JSON from the text
            try:
                json_str = find_json_in_str(tool_result_full)
                if not json_str:
                    logger.error(f"Failed to extract JSON from result: {tool_result_full[:100]}...")
                    return []

                json_data = json.loads(json_str, strict=False)
            except json.JSONDecodeError as e:
                logger.error(
                    f"Failed to parse JSON: {e}, original content: {tool_result_full[:100]}..."
                )
                return []
        result = []
        try:
            if isinstance(json_data, list):
                result = [cls(**item) for item in json_data]
            elif isinstance(json_data, dict):
                result = [cls(**json_data)]
            else:
                raise ValueError(f"Invalid JSON data: {tool_result_full}")
        except (TypeError, ValueError) as e:
            logger.error(f"Invalid JSON data: {tool_result_full}")
            return []

        if save and len(result) > 0:
            for rule_config in result:
                save_rule_config(source=source, rule_config=rule_config)

        return result


class AgentConfig(BaseModel):
    """
    Agent configuration
    """
    name: str
    meta: AgentMeta
    system_prompt: str
    rules: List[RuleConfig]
    max_concurrent_tasks: int = 1
    timeout_detection_sec: int = 180
    stream_registration_timeout: int = 1
    



import asyncio
import uuid
import traceback
import loguru
from loguru._logger import Logger
from typing import Dict, Any, AsyncGenerator
from datetime import datetime

from .task_state import TaskInfo


class OutputChunk:
    """Single output chunk, containing content and corresponding metadata reference"""

    def __init__(self, content: str, task_info: TaskInfo):
        self.content = content
        self.task_info = task_info

    def dict(self) -> Dict[str, Any]:
        """Return a serializable dictionary representation"""
        return {"content": self.content, "task_info": self.task_info.to_dict()}

class OutputStreamManager:
    """Output stream manager, responsible for managing and processing the output of multiple asynchronous generators"""

    def __init__(self, logger: Logger, stream_registration_timeout: float = 1.0):
        """
        Initialize the output stream manager
        Args:
            logger: Logger
            stream_registration_timeout: When all known streams have been processed, the timeout for waiting for new stream registrations (seconds)
        """
        self._logger: Logger = logger
        # _stream_queue stores dictionaries:
        # {"stream_id": str, "generator": AsyncGenerator, "task_info": TaskInfo}
        self._stream_queue = asyncio.Queue()
        self._stream_registration_timeout = stream_registration_timeout
        self._registered_stream_count = 0 # Total number of streams ever registered
        self._streams_processed_count = 0
        self._stream_exhausted = False

    def register_stream(
        self, async_generator: AsyncGenerator[str, None], task_info: TaskInfo
    ) -> str:
        """
        Register a new output stream. The stream will be processed in the order of registration.

        Args:
            async_generator: Asynchronous generator
            task_info: Task information

        Returns:
            stream_id: Stream ID
        """
        stream_id = str(uuid.uuid4())
        # Increment count *before* putting, to signal intent if get_output_stream is checking.
        self._registered_stream_count += 1
        self._stream_queue.put_nowait(
            {
                "stream_id": stream_id,
                "generator": async_generator,
                "task_info": task_info,
            }
        )
        self._logger.info(
            f"Registered stream {stream_id} for task {task_info.task_id}. "
            f"Queue size: {self._stream_queue.qsize()}, Total registered: {self._registered_stream_count}"
        )
        self._stream_exhausted = False
        return stream_id

    async def get_output_stream(self) -> AsyncGenerator[OutputChunk, None]:
        """
        Get and process all output streams in the order of registration.

        This method will sequentially take each registered stream from the queue and fully consume
        all its output before processing the next stream. When all registered streams have been processed,
        and no new streams are registered within a certain timeout period, this asynchronous generator ends.
        """
        if self._stream_exhausted:
            self._logger.info("Output stream already exhausted, returning empty generator.")
            return

        while True:
            stream_data: Dict[str, Any] = {} # Ensure it's defined for 'finally' if an early error occurs
            current_stream_id: str = "N/A"

            try:
                if self._streams_processed_count >= self._registered_stream_count:
                    # All streams registered *up to this point* have been processed.
                    # Now, wait for a *new* stream registration with a timeout.
                    self._logger.info(
                        f"All {self._streams_processed_count}/{self._registered_stream_count} known streams processed. "
                        f"Waiting for new stream registrations with timeout ({self._stream_registration_timeout}s)..."
                    )
                    try:
                        stream_data = await asyncio.wait_for(
                            self._stream_queue.get(), timeout=self._stream_registration_timeout
                        )
                        # If we get here, a new stream was registered and added.
                        # _registered_stream_count would have been incremented by register_stream.
                    except asyncio.TimeoutError:
                        self._logger.info(
                            f"Timeout waiting for new stream registrations. "
                            f"Total processed: {self._streams_processed_count}. No new streams detected. Ending output."
                        )
                        break # Exit the main `while True` loop
                    except asyncio.CancelledError:
                        self._logger.info("Output stream task cancelled while waiting for new stream.")
                        break
                else:
                    # We expect more streams based on _registered_stream_count.
                    # Wait (potentially indefinitely if tasks keep registering) for the next one.
                    self._logger.info(
                        f"Waiting for next stream. Processed: {self._streams_processed_count}/{self._registered_stream_count}. "
                        f"Queue size: {self._stream_queue.qsize()}."
                    )
                    try:
                        stream_data = await self._stream_queue.get()
                    except asyncio.CancelledError:
                        self._logger.info("Output stream task cancelled while waiting for stream.")
                        break
                
                # Deconstruct stream data safely
                generator = stream_data.get("generator")
                task_info = stream_data.get("task_info")
                current_stream_id = stream_data.get("stream_id", "UnknownStreamID")

                if not generator or not task_info: # Basic validation
                    self._logger.error(f"Stream {current_stream_id} received with invalid data (missing generator or task_info). Skipping.")
                    # This stream is considered "processed" to advance past it.
                else:
                    self._logger.info(f"Processing stream {current_stream_id} for task {task_info.task_id}")
                    try:
                        async for content in generator:
                            task_info.update_at = datetime.now() # Update timestamp on activity
                            chunk = OutputChunk(content=content, task_info=task_info)
                            yield chunk
                        self._logger.info(f"Stream {current_stream_id} (Task: {task_info.task_id}) exhausted normally.")
                    except Exception as e:
                        self._logger.error(f"Error processing stream {current_stream_id} (Task: {task_info.task_id}): {e}, {traceback.format_exc()}")
                        if task_info: # Ensure task_info is available to create an error chunk
                            try:
                                error_chunk = OutputChunk(
                                    content=f"Error in stream {current_stream_id}: {str(e)}",
                                    task_info=task_info
                                )
                                yield error_chunk
                            except Exception as yield_error:
                                self._logger.error(f"Failed to yield error chunk for stream {current_stream_id}: {yield_error}")
                        # Stream processing ends here due to error.
            
            except asyncio.CancelledError:
                self._logger.info(f"Output stream task cancelled during processing cycle (around stream {current_stream_id}).")
                break
            except Exception as e: # Catch unexpected errors in the manager's loop logic itself
                self._logger.error(f"Unexpected error in get_output_stream main loop (around stream {current_stream_id}): {e}, {traceback.format_exc()}")
                break # Critical error in manager logic, safer to stop.
            finally:
                # This block ensures that for every item attempted to be processed (even if it failed or was invalid),
                # we mark it as "done" with the queue and increment our processed counter.
                if stream_data: # stream_data would be populated if .get() succeeded
                    self._stream_queue.task_done()
                    self._streams_processed_count += 1
                # self._logger.debug(f"Incremented processed_streams_count to {self._streams_processed_count} after handling item associated with {current_stream_id}.")

        self._logger.info(
            f"Output stream finished. Total streams processed: {self._streams_processed_count}. "
            f"Final registered count: {self._registered_stream_count}."
        )
        self._stream_exhausted = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Agent core class
"""

import loguru
from loguru._logger import Logger
from typing import AsyncGenerator, Optional

from .agent_config_engine import AgentConfigEngine

from .rule_config import AgentConfig
from .utils import get_workspace_root, get_container_workspace
from .context_manager import ContextManager
from .workspace_manager import WorkspaceManager
from .dispatcher import DynamicDispatcher
from .output_stream_manager import OutputChunk



class Agent:
    """Agent core class"""

    def __init__(self, config_dir: str, sid: str):
        """
        Initialize Agent

        Args:
            config_dir: Base path of prompt configuration files
            sid: Session ID
        """
        # FIXME: logger use sid
        self._logger = loguru.logger
        
        if not config_dir or not config_dir.strip():
            raise ValueError("config_dir must be provided")

        self._logger.info(f"Agent init with config_dir: {config_dir}, sid: {sid}")
        self._config: AgentConfig = AgentConfigEngine.load(config_dir)
        self._workspace_manager = WorkspaceManager(
            root=get_workspace_root(sid),
            container_workspace=get_container_workspace(sid),
        )
        self._context_manager = ContextManager(logger=self._logger)
        
        self._context_manager.set_system_prompt(self._config.system_prompt)
        self._dispatcher = DynamicDispatcher(
            initial_rules=self._config.rules,
            workspace_manager=self._workspace_manager,
            context_manager=self._context_manager,
            max_concurrent_tasks=self._config.max_concurrent_tasks,
            timeout_detection_sec=self._config.timeout_detection_sec,
            stream_registration_timeout=self._config.stream_registration_timeout,
            logger=self._logger,
        )

    async def run_dynamic(self, user_input: str, image_url: Optional[str] = None, image_url_detail: Optional[str] = None) -> AsyncGenerator[OutputChunk, None]:
        """
        Run Agent's dynamic conversation flow

        Args:
            user_input: User input
            image_url: Image URL
            image_url_detail: Image URL detail

        Yields:
            str: Generated response chunk
        """
        self._logger.info(f"Agent run_dynamic started with input: {user_input}")

        await self._dispatcher.add_user_input(user_input, image_url=image_url, image_url_detail=image_url_detail)

        async for chunk in self._dispatcher.get_output_stream():
            yield chunk

        self._logger.info("Agent run_dynamic finished.")
        # TODO: Consider a shutdown for the dispatcher if the agent instance is not reused
        # await self._dispatcher.shutdown()



from typing import Dict, Any, AsyncGenerator
import asyncio
import time
import loguru
from loguru._logger import Logger
import json
import traceback

from .rule_registery import RuleRegistry
from .context_manager import ContextManager
from .llm.llm_base import TextResult, AIContext
from .tools.base import ToolCenter
from .rule_config import RuleConfig
from .task_state import TaskInfo


class MaxRetryError(Exception):
    """Tool call reached maximum retry count"""

    error_msg: str

    def __init__(self, error_msg: str):
        self.error_msg = error_msg
        super().__init__(error_msg)


class TaskExecutor:
    """Task executor, responsible for executing tasks and handling results"""

    def __init__(
        self,
        rule_registry: RuleRegistry,
        context_manager: ContextManager,
        logger: Logger,
        # event_bus: EventBus, # 响应复杂情况时使用
    ):
        self._rule_registry: RuleRegistry = rule_registry
        self._context_manager: ContextManager = context_manager
        self._logger: Logger = logger
        self._executing_tasks: Dict[str, asyncio.Task] = {}

    async def _prepare_tool_params(self, task_info: TaskInfo) -> Dict[str, Any]:
        """Prepare tool call parameters"""
        rule_config: RuleConfig = task_info.rule_config

        tool_params = {}

        # Get parameters from context
        ctx_keys = rule_config.tool_params.get("ctx", [])
        for key in ctx_keys:
            tool_params[key] = self._context_manager.get_context().get(key)

        # Get parameters from rule config
        config_keys = rule_config.tool_params.get("config", [])
        for key in config_keys:
            tool_params[key] = getattr(rule_config, key)

        # Add extra parameters
        extra_params = rule_config.tool_params.get("extra", {}) or {}
        tool_params.update(extra_params)

        # Add context and rule config
        tool_params["ctx"] = self._context_manager.get_context()
        tool_params["ctx_manager"] = self._context_manager
        tool_params["rule_config"] = rule_config

        return tool_params

    async def _handle_tool_result(
        self, task_info: TaskInfo, response_full: str
    ) -> None:
        """Handle tool call result"""
        rule_config: RuleConfig = task_info.rule_config

        if rule_config.tool_result_target == "DIRECT_RETURN":
            # Save to conversation history
            await self._context_manager.emit_and_append_to_history(
                TextResult(text=response_full)
            )

        elif rule_config.tool_result_target == "AS_CONTEXT":
            # Save to context
            self._context_manager.get_context().set(
                rule_config.tool_result_key, response_full
            )
            await self._context_manager.emit_and_append_to_history(
                AIContext(key=rule_config.tool_result_key, context=response_full)
            )

        elif rule_config.tool_result_target == "NEW_RULES":
            # Parse and generate new rules
            new_rule_configs = RuleConfig.parse_and_gen(
                source=rule_config.name,
                tool_result_full=response_full,
                save=True,
            )
            await self._context_manager.emit_and_append_to_history(
                AIContext(context=response_full)
            )
            for new_cfg in new_rule_configs:
                new_cfg.auto_generated = True
                self._context_manager.emit_task_generate_new_rule(
                    task_info, new_cfg, immediate=True
                )

    async def _execute_task(self, task_info: TaskInfo) -> AsyncGenerator[str, None]:
        """Internal implementation of task execution"""
        start_time = time.time()
        task_id = task_info.task_id
        rule_config: RuleConfig = task_info.rule_config
        response_full = ""
        error_msg = None

        async def emit_error(msg: str, error_type: str = "failed") -> None:
            """Helper function: send error message and update status"""
            nonlocal error_msg
            error_msg = msg
            yield msg

            execution_time = time.time() - start_time
            if error_type == "cancelled":
                self._context_manager.emit_task_cancelled(task_info, msg)
            else:
                self._context_manager.emit_task_failed(task_info, execution_time, msg)

        try:
            # 1. Prepare tool call parameters
            try:
                tool_params = await self._prepare_tool_params(task_info)
                self._logger.info(
                    f"Task for Rule: {rule_config.name} {task_info.rule_id}: Tool params prepared"
                )
            except Exception as e:
                self._logger.error(
                    f"Error: Failed to prepare parameters for tool {rule_config.tool}: {e}, {traceback.format_exc()}"
                )
                async for chunk in emit_error(
                    f"Error: Failed to prepare parameters for tool {rule_config.tool}: {e}"
                ):
                    yield chunk
                return

            # 2. Execute tool call
            self._context_manager.emit_task_executing(task_info)
            try:
                async for chunk in ToolCenter.run_tool(
                    name=rule_config.tool, **tool_params
                ):
                    if rule_config.tool_result_target == "DIRECT_RETURN":
                        yield chunk
                    response_full += chunk
            except asyncio.CancelledError:
                self._logger.warning(f"Task {task_id}: Tool execution was cancelled")
                async for chunk in emit_error(
                    "Tool execution was cancelled", "cancelled"
                ):
                    yield chunk
                raise  # Re-raise cancellation error
            except Exception as e:
                self._logger.error(
                    f"Task {task_id}: Tool {rule_config.tool} execution failed: err={e}, {traceback.format_exc()}"
                )
                async for chunk in emit_error(
                    f"Error: Tool {rule_config.tool} execution failed: {e}"
                ):
                    yield chunk
                return

            # 3. Handle tool call result
            if not response_full:
                self._logger.error(
                    f"Error: Tool {rule_config.tool}, task {task_id} returned empty result"
                )
                async for chunk in emit_error(
                    f"Error: Tool {rule_config.tool}, task {task_id} returned empty result"
                ):
                    yield chunk
                return

            try:
                await self._handle_tool_result(task_info, response_full)
                self._logger.info(f"Task {task_id}: Tool result handled")
            except json.JSONDecodeError as e:
                self._logger.error(
                    f"Error: Tool {rule_config.tool}, task {task_id} result handling failed (invalid format): {e}, {traceback.format_exc()}"
                )
                async for chunk in emit_error(
                    f"Error: Tool {rule_config.tool}, task {task_id} result handling failed (invalid format): {e}"
                ):
                    yield chunk
                return
            except Exception as e:
                self._logger.error(
                    f"Error: Tool {rule_config.tool}, task {task_id} result handling failed: err={e}, {traceback.format_exc()}"
                )
                async for chunk in emit_error(
                    f"Error: Tool {rule_config.tool}, task {task_id} result handling failed: {e}"
                ):
                    yield chunk
                return

            # Task completed successfully
            execution_time = time.time() - start_time
            self._logger.info(
                f"Task {task_id} completed successfully in {execution_time:.2f}s"
            )
            self._context_manager.emit_task_completed(
                task_info, execution_time, response_full
            )

        except asyncio.CancelledError:
            self._logger.warning(f"Task {task_id} was cancelled")
            async for chunk in emit_error(f"Task {task_id} was cancelled", "cancelled"):
                yield chunk
            raise  # Re-raise cancellation error

        except Exception as e:
            self._logger.error(
                f"Task {task_id}: Unexpected error in execute_task: err={e}, {traceback.format_exc()}"
            )
            async for chunk in emit_error(
                f"Task {task_id}: Unexpected error in execute_task: {e}"
            ):
                yield chunk

    async def run_task(self, task_info: TaskInfo) -> AsyncGenerator[str, None]:
        """Run task and support cancellation

        This method creates an async task and stores it in the _executing_tasks dictionary
        to allow cancellation of the task.
        """
        task_id = task_info.task_id

        # Create an internal function to wrap execute_task and clean up the task on completion
        async def _task_wrapper():
            try:
                async for chunk in self._execute_task(task_info):
                    yield chunk
            finally:
                # Remove task from dictionary on completion
                if task_id in self._executing_tasks:
                    del self._executing_tasks[task_id]

        # Create async generator
        gen = _task_wrapper().__aiter__()

        # Create a task to run the generator
        task = asyncio.create_task(gen.__anext__())
        self._executing_tasks[task_id] = task

        # Return generator results
        try:
            while True:
                try:
                    # Wait for current chunk
                    result = await task
                    yield result

                    # Create next chunk task
                    task = asyncio.create_task(gen.__anext__())
                    self._executing_tasks[task_id] = task
                except StopAsyncIteration:
                    # Generator completed
                    break
        finally:
            # Ensure task is removed from dictionary
            if task_id in self._executing_tasks:
                del self._executing_tasks[task_id]

    async def cancel_task(self, task_id: str) -> None:
        """Cancel task"""
        self._logger.info(f"Cancelling task {task_id}")
        if task_id in self._executing_tasks:
            task = self._executing_tasks[task_id]
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass
            # The task will be removed from the dictionary in the finally block of run_task



import loguru
from loguru._logger import Logger
import asyncio
from typing import List, AsyncGenerator, Optional
from .workspace_manager import WorkspaceManager
from .context_manager import ContextManager
from .event_bus import EventType, RealTimeEvent
from .llm.llm_base import TextPrompt
from .task_manager import TaskManager
from .rule_registery import RuleRegistry
from .task_executor import TaskExecutor
from .output_stream_manager import OutputStreamManager, OutputChunk
from .rule_config import RuleConfig


class DynamicDispatcher:
    """
    Dynamic dispatcher for task management and rule execution
    """

    def __init__(
        self,
        initial_rules: List[RuleConfig],
        workspace_manager: WorkspaceManager,
        context_manager: ContextManager,
        max_concurrent_tasks: int,
        timeout_detection_sec: int,
        stream_registration_timeout: int,
        logger: Logger,
    ):
        self._rules = initial_rules
        self._workspace_manager = workspace_manager
        self._context_manager = context_manager
        self._logger = logger
        self._rule_registry = RuleRegistry()
        self._task_executor = TaskExecutor(
            self._rule_registry, self._context_manager, self._logger
        )
        self._output_manager = OutputStreamManager(logger=logger, stream_registration_timeout=stream_registration_timeout)
        self._task_manager = TaskManager(
            self._context_manager,
            executor=self._task_executor,
            output_manager=self._output_manager,
            logger=self._logger,
            max_concurrent_tasks=max_concurrent_tasks,
            timeout_detection_sec=timeout_detection_sec,
        )
        # Initialize rules
        if initial_rules:
            for rule_config in initial_rules:
                self.add_rule(rule_config)

        # Subscribe to context changes
        self._context_manager.subscribe(
            EventType.CONTEXT_CHANGED, self._handle_ctx_changed
        )
        self._context_manager.subscribe(
            EventType.NEW_RULE_GENERATED, self._handle_new_rule_generated
        )
        self._context_manager.subscribe(
            EventType.NEW_RULE_GENERATED, self._handle_new_rule_generated
        )
        # TODO: more events

    async def _handle_ctx_changed(self, event: RealTimeEvent) -> None:
        """
        Handle context changes
        """
        changed_key = event.data["key"]
        new_value = event.data["value"]
        old_value = event.data["old_value"]
        self._logger.info(
            f"Context changed: key='{changed_key}', value='{new_value}', old_value='{old_value}'"
        )
        # Get all rules depending on this key
        dependent_rules = self._rule_registry.get_rules_for_key(changed_key)
        self._logger.info(
            f"Found {len(dependent_rules)} rules depending on key '{changed_key}'"
        )
        # Check each rule and create task
        for rule_id in dependent_rules:
            await self._check_and_create_task(rule_id)

    async def _handle_new_rule_generated(self, event: RealTimeEvent):
        event_data = event.data
        task_id = event_data["task_id"]
        rule_config = event_data["rule_config"]
        immediate = event_data.get("immediate", False)
        self._logger.info(
            f"New rule generated: {rule_config} by task {task_id}, immediate={immediate}"
        )
        await self.add_rule(rule_config=rule_config, immediate=immediate)

    async def add_user_input(self, user_input: str, image_url: Optional[str] = None, image_url_detail: Optional[str] = None):
        self._logger.info(f"Received user input: {user_input}")
        text_prompt = TextPrompt(text=user_input, image_url=image_url, image_url_detail=image_url_detail)
        await self._context_manager.emit_and_append_to_history(text_prompt)

    def add_rule(self, rule_config: RuleConfig, immediate: bool = False) -> str:
        """Add new rule"""
        rule_id = self._rule_registry.register_rule(rule_config)
        # If immediate check is needed, check conditions and create tasks if satisfied
        if immediate:
            asyncio.create_task(self._check_and_create_task(rule_id))

        return rule_id

    async def _check_and_create_task(self, rule_id: str) -> None:
        """Check rule conditions and create tasks if satisfied"""
        # Check if there are any active tasks for this rule
        rule_tasks = self._task_manager.get_tasks_by_rule(rule_id)
        active_tasks = [
            task_id
            for task_id in rule_tasks
            if self._task_manager.is_task_executing(task_id)
        ]

        if active_tasks:
            self._logger.info(
                f"Rule {rule_id} already has active tasks. Skipping task creation."
            )
            return

        # Check rule conditions
        if not self._rule_registry.check_rule_condition(
            rule_id, self._context_manager.get_context()
        ):
            self._logger.info(
                f"Rule {rule_id} does not meet the condition. Skipping task creation."
            )
            return

        # Condition satisfied, create new task
        rule_config = self._rule_registry.get_rule(rule_id)
        task_id = await self._task_manager.create_task_and_schedule(
            rule_id, rule_config
        )
        self._logger.info(
            f"Created new task {task_id} for rule {rule_id}, priority={rule_config.priority}"
        )

    async def get_output_stream(self) -> AsyncGenerator[OutputChunk, None]:
        """Provide final output as an async generator"""
        try:
            # Continuously check if there are any active tasks or pending tasks
            while True:
                # Check if there is any output available
                has_output = False
                async for chunk in self._output_manager.get_output_stream():
                    has_output = True
                    yield chunk

                # Check if all tasks are completed
                active_tasks = self._task_manager.get_active_tasks()
                active_task_count = len(active_tasks)
                # If there are no active tasks and no output, end the stream
                if not active_tasks and not has_output and active_task_count == 0:
                    self._logger.info("All tasks are completed and output stream is exhausted, ending output stream")
                    break
                self._logger.info(f"Waiting for output stream, current active task count: {active_task_count}")
                if active_task_count > 0:
                    await asyncio.sleep(0.2)
                else:
                    break
        finally:
            self._logger.info("Output stream ended")

    async def shutdown(self) -> None:
        """Shutdown dispatcher and clean up resources"""
        self._logger.info("Shutting down Dispatcher...")
        # Close task manager
        await self._task_manager.shutdown()
        self._logger.info("Dispatcher shutdown complete")



from typing import Deque, List, Dict, Any, Callable
from pydantic import BaseModel
import asyncio
import enum
import loguru
from loguru._logger import Logger
from loguru._logger import Logger
import threading


class EventType(str, enum.Enum):
    USER_MESSAGE = "user_message"  # User input
    TOOL_CALL = "tool_call"  # ToolCall
    TOOL_RESULT = "tool_result"  # ToolResult
    TOKEN_USAGE = "token_usage"  # Token usage
    AGENT_RESPONSE = "agent_response"  # AgentResponse
    CONTEXT_CHANGED = "context_changed"  # Internal system-triggered context change
    NEW_RULE_GENERATED = "new_rule_generated"  # New rule generated
    RULE_TASK_EXECUTING = "rule_task_executing"  # RuleTask is executing
    RULE_TASK_COMPLETED = "rule_task_completed"  # RuleTask completed
    RULE_TASK_FAILED = "rule_task_failed"  # RuleTask failed
    RULE_TASK_CANCELLED = "rule_task_cancelled"  # RuleTask cancelled


class RealTimeEvent(BaseModel):
    type: EventType
    data: Dict[str, Any]


class EventBus:
    """
    Event bus for internal event passing within Agent
    """

    def __init__(self, logger: Logger):
        self._logger = logger
        self._queue = asyncio.Queue()
        # Event subscribers mapping: event type -> callback function list
        self._subscribers: Dict[EventType, List[Callable[..., Any]]] = {}
        # Event history: event type -> recent event data queue
        self._event_history: Dict[EventType, Deque] = {}

        # Event type lock mapping: event type -> lock
        self._event_locks: Dict[EventType, threading.Lock] = {}
        # Dictionary lock, used to protect the modification of _subscribers and _event_history dictionaries
        self._dict_lock = threading.Lock()

    def _get_event_lock(self, event_type: EventType) -> threading.Lock:
        """
        Get the lock corresponding to the event type, create if it does not exist

        Args:
            event_type: Event type

        Returns:
            Lock corresponding to the event type
        """
        with self._dict_lock:
            if event_type not in self._event_locks:
                self._event_locks[event_type] = threading.Lock()
            return self._event_locks[event_type]

    def emit(self, event: RealTimeEvent):
        """
        Emit an event
        """
        self._logger.debug(f"Received event {event.type}, data {event.model_dump()}")
        self._queue.put_nowait(event)
        # Record event history
        subscribers = []
        event_lock = self._get_event_lock(event.type)
        with event_lock:
            if event.type not in self._event_history:
                self._event_history[event.type] = Deque(maxlen=1024)
            self._event_history[event.type].append(event.model_dump())
            subscribers = self._subscribers.get(event.type, [])
        # Send event to subscribers
        for callback in subscribers:
            self._logger.debug(f"Event {event.type} sent to subscriber {callback.__name__}")
            asyncio.create_task(callback(event))

    def subscribe(
        self, event_type: EventType, callback: Callable[[RealTimeEvent], None]
    ):
        """
        Subscribe to an event
        """
        self._logger.debug(f"Received event subscription {event_type}, callback {callback.__name__}")
        event_lock = self._get_event_lock(event_type)
        with event_lock:
            self._subscribers.setdefault(event_type, []).append(callback)

    def unsubscribe(
        self, event_type: EventType, callback: Callable[[RealTimeEvent], None]
    ):
        """
        Unsubscribe from an event
        """
        self._logger.debug(
            f"Received event unsubscription {event_type}, callback {callback.__name__}"
        )
        event_lock = self._get_event_lock(event_type)
        with event_lock:
            if event_type not in self._subscribers:
                self._logger.debug(f"Event {event_type} has no subscribers, unsubscribe failed")
            else:
                self._subscribers[event_type].remove(callback)
                if not self._subscribers[event_type]:
                    del self._subscribers[event_type]
                    self._logger.debug(f"Event {event_type} has no subscribers, cleaned up")



from typing import Any, Dict
from pathlib import Path


def get_workspace_root(sid: str) -> Path:
    return Path(f"./akr-agent/{sid}")


def get_container_workspace(sid: str) -> Path:
    return Path(f"./akr-agent/{sid}/container")


def get_nested(data: Dict, keys: str, default: Any = None) -> Any:
    """
    Access nested dictionary keys using dot notation.

    Args:
        data: The dictionary to access
        keys: The dot notation key path, e.g. "a.b.c"
        default: The default value to return if the key does not exist

    Returns:
        The value found or the default value
    """
    if not keys:
        return data

    key_list = keys.split(".")
    current = data

    for key in key_list:
        if isinstance(current, dict):
            current = current.get(key, default)
        elif isinstance(current, list):
            try:
                idx = int(key)
                if 0 <= idx < len(current):
                    current = current[idx]
                else:
                    return default
            except (ValueError, IndexError):
                return default
        else:
            return default

        if current is default:
            break

    return current


def set_nested(data: Dict, keys: str, value: Any) -> None:
    """
    Set nested dictionary keys using dot notation.

    Args:
        data: The dictionary to modify
        keys: The dot notation key path, e.g. "a.b.c"
        value: The value to set
    """
    if not keys:
        return

    key_list = keys.split(".")
    current = data

    for i, key in enumerate(key_list[:-1]):
        if key not in current or not isinstance(current[key], dict):
            current[key] = {}
        current = current[key]

    current[key_list[-1]] = value



"""
OpenAI LLM client implementation
"""

import asyncio
from typing import Any, AsyncGenerator, Dict, Optional, List, Callable
import json
import traceback
import openai
from openai import AsyncOpenAI
from openai.types.chat import ChatCompletionChunk
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
)

from .base import LLMClient
from .llm_base import ToolCall, ToolResult, TokenUsage

from loguru import logger


class OpenAIClient(LLMClient):
    """
    OpenAI API client implementation (support tool_calls)
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        model: str = "gpt-4o-mini",  # suggest use new model support tool_calls
        temperature: float = 0.7,
        max_tokens: Optional[int] = None,
        **kwargs,
    ):
        """
        Initialize OpenAI client

        Args:
            api_key: OpenAI API key, if None then use environment variable
            model: Model name
            temperature: Temperature parameter
            max_tokens: Maximum tokens
            **kwargs: Other OpenAI API parameters (e.g. base_url, timeout, etc.)
        """
        self.client = AsyncOpenAI(
            api_key=api_key, base_url=base_url, **kwargs.pop("client_args", {})
        )  # pass additional parameters to client
        self.model = model
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.extra_params = kwargs  # other parameters passed to completions.create
        logger.info(f"Initialize OpenAI client, model: {model}")

    async def invoke_stream(
        self,
        system_prompt: str,
        user_input: str,
        messages: Optional[List[Dict[str, Any]]] = None,
        run_tool_func: Optional[Callable[[str, str], Any]] = None,
        **kwargs,
    ) -> AsyncGenerator[str, None]:
        """
        Streamly invoke OpenAI API and return response stream, support tool_calls.

        Args:
            system_prompt: System prompt (only used to build initial messages on first call or when messages is None)
            user_input: User input (only used to build initial messages on first call or when messages is None)
            messages: Optional, preset messages list.
            run_tool_func: Optional async function to execute tool calls. Signature should be: async def run_tool(tool_name: str, tool_args: str) -> Any
            **kwargs: Override default parameters or pass additional parameters (e.g. tools, tool_choice)
                - ctx_manager: Context manager, used to output assistant messages

        Yields:
            Response fragment (str)
        """
        # prepare API call parameters
        # note: kwargs passed to _prepare_params, it will merge instance attributes and these runtime parameters
        api_params = {
            "model": self.model,
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
        }
        api_params.update(
            self._prepare_params(
                system_prompt=system_prompt,  # pass to _prepare_params for possible initial message building logic
                user_input=user_input,  # same as above
                messages=messages,  # most important: pass current message history
                **kwargs,  # include tools, tool_choice, etc
            )
        )
        api_params["stream"] = True
        api_params["stream_options"] = {
            "include_usage": True,
        }
        if "ctx_manager" in kwargs:
            from ..context_manager import ContextManager

            ctx_manager: ContextManager = kwargs.get("ctx_manager")
        else:
            ctx_manager = None

        max_retries = 3
        retry_count = 0
        backoff_factor = 2

        while retry_count <= max_retries:
            try:
                logger.info(f"OPENAI invoke stream params: {api_params}")
                response_stream: AsyncGenerator[ChatCompletionChunk, None] = (
                    await self.client.chat.completions.create(**api_params)
                )

                # used to accumulate tool_calls data in current LLM response
                # key: tool_call_id, value: {"id": ..., "type": "function", "function_name": ..., "function_arguments": ...}
                active_tool_calls_data: Dict[str, Dict[str, str]] = {}
                tool_calls = []
                # used to accumulate text content in current LLM response (if LLM speaks before requesting tool calls)
                # current_assistant_content_parts: List[str] = []
                token_usage = None
                async for chunk in response_stream:
                    if chunk.usage and token_usage is None:
                        # token usage statistics
                        token_usage = chunk.usage
                        total_tokens = chunk.usage.total_tokens
                        prompt_tokens = chunk.usage.prompt_tokens
                        completion_tokens = chunk.usage.completion_tokens
                        completion_tokens_details = (
                            chunk.usage.completion_tokens_details
                        )
                        prompt_tokens_details = chunk.usage.prompt_tokens_details
                        if ctx_manager is not None:
                            await ctx_manager.emit_and_append_to_history(
                                TokenUsage(
                                    role="assistant",
                                    model=self.model,
                                    total_tokens=total_tokens,
                                    prompt_tokens=prompt_tokens,
                                    completion_tokens=completion_tokens,
                                    completion_tokens_details=(
                                        completion_tokens_details.to_dict()
                                        if completion_tokens_details
                                        else None
                                    ),
                                    prompt_tokens_details=(
                                        prompt_tokens_details.to_dict()
                                        if prompt_tokens_details
                                        else None
                                    ),
                                )
                            )
                        logger.info(
                            f"Llm Call Token cost: {chunk.usage.model_dump_json()}"
                        )

                    if not chunk.choices:
                        continue
                    delta = chunk.choices[0].delta
                    finish_reason = chunk.choices[0].finish_reason

                    # 1. process text content
                    if delta and delta.content:
                        # current_assistant_content_parts.append(delta.content)
                        yield delta.content

                    # 2. process tool_calls
                    if delta and delta.tool_calls:
                        tc_chunk_list = delta.tool_calls
                        for tc_chunk in tc_chunk_list:
                            logger.debug(f"PROCESS TOOL CALL CHUNK, {tc_chunk}")
                            if len(tool_calls) <= tc_chunk.index:
                                tool_calls.append(
                                    {
                                        "id": "",
                                        "type": "function",
                                        "function": {"name": "", "arguments": ""},
                                    }
                                )
                            tc = tool_calls[tc_chunk.index]

                            if tc_chunk.id:
                                tc["id"] += tc_chunk.id
                            if tc_chunk.function.name:
                                tc["function"]["name"] += tc_chunk.function.name
                            if tc_chunk.function.arguments:
                                tc["function"][
                                    "arguments"
                                ] += tc_chunk.function.arguments

                    # 3. when LLM indicates tool calls are finished (or stream naturally ends)
                    if finish_reason == "tool_calls":
                        if len(tool_calls) == 0:
                            logger.warning(
                                "finish_reason is 'tool_calls' but no tool call data collected."
                            )
                            # This should not happen, but just in case
                            break

                        if not run_tool_func:
                            yield "\n[Error: LLM requests tool calls, but 'run_tool_func' is not provided to execute them.]\n"
                            logger.error(
                                "LLM requests tool calls, but 'run_tool_func' is not provided."
                            )
                            return  # end generation

                        # 3.1 build assistant message history (include tool call requests)
                        # streamed_content = "".join(current_assistant_content_parts)
                        # current_assistant_content_parts.clear() # clear for next round (if recursive)

                        logger.info(f"llm finsih with tool_calls, {tool_calls}")

                        assistant_tool_calls_list: List[
                            ChatCompletionMessageToolCall
                        ] = []
                        tools_to_execute_details = []  # used to execute tools

                        for tool_call in tool_calls:
                            tc_id = tool_call["id"]
                            function_name = tool_call["function"]["name"]
                            args_json = tool_call["function"]["arguments"]
                            function_args = json.loads(args_json)
                            logger.info(
                                f"parse toolcall, {tc_id}, {function_name}, {args_json}, {function_args}"
                            )
                            if (
                                function_name and tc_id
                            ):  # Parameters may be empty strings, but name and ID must be present
                                assistant_tool_calls_list.append(
                                    ChatCompletionMessageToolCall(
                                        id=tc_id,
                                        type="function",  # OpenAI currently only supports function type
                                        function={
                                            "name": function_name,
                                            "arguments": args_json,
                                        },
                                    )
                                )
                                tools_to_execute_details.append(
                                    {
                                        "id": tc_id,
                                        "name": function_name,
                                        "arguments": function_args,
                                    }
                                )
                            else:
                                logger.warning(
                                    f"collect tool call data incomplete, ID {tc_id}: {tool_call}"
                                )

                        if not assistant_tool_calls_list:
                            logger.error(
                                "finish_reason='tool_calls' but no valid tool calls to execute."
                            )
                            yield "\n[error: LLM requests tool calls, but unable to parse valid tool information.]\n"
                            return
                        logger.info(
                            f"add new Tool Call to Current Messages, tool_calls={assistant_tool_calls_list}"
                        )
                        current_messages.append(
                            {
                                "role": "assistant",
                                "content": None,  # Text content has already been yielded
                                "tool_calls": [
                                    # SDK v1.x ChatCompletionMessageToolCall is not directly JSON serializable for message history
                                    # We need to convert them to dicts if we were to manually build this
                                    # However, the openai library handles this internally if we pass the objects.
                                    # For clarity and if current_messages is used outside this specific SDK context,
                                    # converting to dict structure matching API spec is safer.
                                    {
                                        "id": tc.id,
                                        "type": tc.type,
                                        "function": {
                                            "name": tc.function.name,
                                            "arguments": tc.function.arguments,
                                        },
                                    }
                                    for tc in assistant_tool_calls_list
                                ],
                            }
                        )
                        if ctx_manager:
                            for tc in assistant_tool_calls_list:
                                await ctx_manager.emit_and_append_to_history(
                                    ToolCall(
                                        tool_call_id=tc.id,
                                        tool_name=tc.function.name,
                                        tool_input=tc.function.arguments,
                                    )
                                )

                        # 3.2 execute tools
                        tool_results_messages: List[Dict[str, Any]] = []
                        # can consider using asyncio.gather to parallel execute independent tools
                        for tool_data in tools_to_execute_details:
                            tool_name = tool_data["name"]
                            tool_args = tool_data["arguments"]
                            tool_call_id = tool_data["id"]

                            logger.info(
                                f"invoke tool: {tool_name}, args: {tool_args}, id: {tool_call_id}"
                            )
                            try:
                                # call run_tool_func, it may return coroutine or async generator
                                tool_output = run_tool_func(tool_name, **tool_args)

                                tool_result_content = ""
                                if isinstance(tool_output, AsyncGenerator):
                                    # if it's async generator, accumulate all outputs
                                    accumulated_parts = []
                                    async for part in tool_output:
                                        accumulated_parts.append(str(part))
                                    tool_result_content = "".join(accumulated_parts)
                                    if (
                                        not tool_result_content
                                        and not accumulated_parts
                                    ):  # distinguish empty string and no yield at all
                                        tool_result_content = (
                                            "[tool executed, but no content generated]"
                                        )
                                elif asyncio.iscoroutine(tool_output):
                                    # if it's coroutine, await directly
                                    tool_result_content = await tool_output
                                else:
                                    # if it's sync function return value (despite run_tool_func expects async)
                                    # or other unexpected type, try to convert to string
                                    logger.warning(
                                        f"tool {tool_name} returned unexpected type: {type(tool_output)}. Trying to convert to string."
                                    )
                                    tool_result_content = str(tool_output)

                                logger.info(
                                    f"tool {tool_name} (ID: {tool_call_id}) result: {tool_result_content}"
                                )
                                tool_results_messages.append(
                                    {
                                        "role": "tool",
                                        "tool_call_id": tool_call_id,
                                        "name": tool_name,
                                        "content": str(
                                            tool_result_content
                                        ),  # result must be string
                                    }
                                )
                            except Exception as e:
                                logger.error(
                                    f"tool {tool_name} (ID: {tool_call_id}) execution failed: err={e}, trace={traceback.format_exc()}"
                                )
                                tool_results_messages.append(
                                    {
                                        "role": "tool",
                                        "tool_call_id": tool_call_id,
                                        "name": tool_name,
                                        "content": f"error: failed to execute tool {tool_name}: {str(e)}",
                                    }
                                )

                        current_messages.extend(tool_results_messages)
                        if ctx_manager:
                            for tool_result in tool_results_messages:
                                await ctx_manager.emit_and_append_to_history(
                                    ToolResult(
                                        tool_call_id=tool_result["tool_call_id"],
                                        tool_name=tool_result["name"],
                                        tool_output=tool_result["content"],
                                    )
                                )

                        # 3.3 Recursive call, continue getting LLM response
                        # Clear active_tool_calls_data for next LLM response (although a new instance will be created in recursive call)
                        active_tool_calls_data.clear()

                        async for content_piece in self.invoke_stream(
                            system_prompt=system_prompt,  # These are used for _prepare_params default parameters in recursive call
                            user_input=user_input,  # Actual history is controlled by messages
                            messages=current_messages,  # Pass updated complete message history
                            run_tool_func=run_tool_func,  # Pass tool execution function
                            **kwargs,  # Pass other parameters like tools, tool_choice
                        ):
                            yield content_piece
                        return  # End this round of invoke_stream, because recursive call has handled the rest

                # 4. If the stream ends normally (finish_reason='stop', 'length', etc.) and there are no unprocessed tool calls
                if active_tool_calls_data and finish_reason != "tool_calls":
                    # This should not happen, if LLM plans to call tools, finish_reason should be tool_calls
                    logger.warning(
                        f"Stream ended (finish_reason: {finish_reason}), but there are still unprocessed tool call data: {active_tool_calls_data}"
                    )
                    yield f"\n[warning: stream ended unexpectedly, there may be incomplete tool call requests: {list(active_tool_calls_data.keys())}]\n"

                break  # Successfully completed or normally ended, exit retry loop

            except openai.RateLimitError as e:
                retry_count += 1
                if retry_count <= max_retries:
                    wait_time = backoff_factor**retry_count
                    logger.warning(
                        f"API rate limit exceeded. Will retry in {wait_time}s. Attempt {retry_count}/{max_retries}"
                    )
                    await asyncio.sleep(wait_time)
                else:
                    logger.error(
                        f"API rate limit exceeded, reached maximum retry count: {e}"
                    )
                    yield "\nError: API rate limit exceeded, please try again later.\n"
                    break
            except openai.AuthenticationError as e:
                logger.error(f"OpenAI API authentication failed: {e}")
                yield "\nError: API authentication failed, please check API key configuration.\n"
                break
            except (openai.APIConnectionError, asyncio.TimeoutError) as e:
                retry_count += 1
                if retry_count <= max_retries:
                    wait_time = backoff_factor**retry_count
                    logger.warning(
                        f"Connection error: {e}. Will retry in {wait_time}s. Attempt {retry_count}/{max_retries}"
                    )
                    await asyncio.sleep(wait_time)
                else:
                    logger.error(
                        f"Failed to connect to OpenAI API, reached maximum retry count: {e}"
                    )
                    yield f"\nError: Failed to connect to OpenAI API: {e}\n"
                    break
            except asyncio.CancelledError:
                logger.info("OpenAI API request cancelled")
                break  # Do not retry
            except Exception as e:
                logger.error(
                    f"Unexpected error in OpenAI API call: err={e}, trace={traceback.format_exc()}"
                )
                yield f"\nError: {str(e)}\n"
                break  # Do not retry unknown error

    def _prepare_params(
        self,
        system_prompt: str,
        user_input: str,
        messages: Optional[List[Dict[str, Any]]] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Prepare API call parameters

        Args:
            system_prompt: System prompt
            user_input: User input
            messages: Current conversation message list
            **kwargs: Runtime parameters, will override instance default settings

        Returns:
            API call parameters dictionary
        """
        current_messages: List[Dict[str, Any]] = []

        current_messages.append({"role": "system", "content": system_prompt})
        if messages is not None:
            current_messages.extend(messages)  # use provided message list copy
        new_input = []
        if kwargs.get("image_url", None) is not None:
            image_url_detail = kwargs.get("image_url_detail", "high")
            new_input.extend(
                [
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": kwargs.get("image_url"),
                            "detail": image_url_detail,
                        },
                    },
                    {"type": "text", "text": user_input},
                ]
            )
        else:
            new_input.append({"type": "text", "text": user_input})
        current_messages.append({"role": "user", "content": new_input})

        logger.debug(f"OpenAI History messages: {current_messages}")

        # Basic parameters, allow being overridden by kwargs
        params: Dict[str, Any] = {
            "model": kwargs.get("model", self.model),
            "temperature": kwargs.get("temperature", self.temperature),
            "messages": current_messages,
        }

        # Optional parameters
        max_tokens_to_use = kwargs.get("max_tokens", self.max_tokens)
        if max_tokens_to_use is not None:
            params["max_tokens"] = max_tokens_to_use

        # Tool related parameters (from kwargs)
        if "tools" in kwargs:
            params["tools"] = kwargs["tools"]
        if (
            "tool_choice" in kwargs
        ):  # e.g., "auto", "none", {"type": "function", "function": {"name": "my_function"}}
            params["tool_choice"] = kwargs["tool_choice"]

        # Merge self.extra_params (parameters not specified in kwargs)
        for key, value in self.extra_params.items():
            if (
                key not in params
            ):  # Avoid overriding parameters already set in kwargs or method fixed settings
                params[key] = value

        if "tools" in params and len(params.get("tools", [])) == 0:
            params.pop("tools")
            if "tool_choice" in params:
                params.pop("tool_choice")
        if "tool_choice" in params and len(params["tool_choice"]) == 0:
            params.pop("tool_choice")

        # Ensure other parameters not explicitly handled in kwargs are also added, allowing complete flexibility
        # But be careful not to override core parameters like 'model', 'messages', 'stream' that have already been set
        # This logic can be refined, but usually self.extra_params and explicit kwargs already cover most cases
        # for key, value in kwargs.items():
        #     if key not in params and key not in ["system_prompt", "prompt", "current_messages", "client_args"]:
        #         params[key] = value
        # The above logic may be too broad, if kwargs contains internal used like system_prompt, should not directly add to params
        # Usually tools, tool_choice, response_format etc. should be explicitly passed to completions.create
        # self.extra_params can be used to store some API parameters that do not change often.

        return params





from .langfun_python_code import PythonCode
from .langfun_correction import CodeWithError, CorrectedCode

def fix_code_error_by_llm(code_with_error: CodeWithError):
    """Fixes the code by using a language model."""
    
    code_text = code_with_error.code
    error = code_with_error.error
    
    prompt = """Given the following code and error:
    Code:
    ```python
    {code}
    ```
    
    Error:
    {error}
    
    Fix the code and return the final corrected code only.
    """
    # TODO: use llm to fix the code
    # result = llm.invoke(prompt.format(code=code_text, error=error))
    
    return CorrectedCode(corrected_code=code_text)

class PythonCodeExecutor:
    def __init__(self):
        pass

    def run(self, code_text: str, **kwargs):
        code = PythonCode(source=code_text)
        result = code(
            sandbox=False,
            timeout=30,
            autofix=3,
            outputs_intermediate=True,
            returns_stdout=False,
            fix_code_error_by_llm=fix_code_error_by_llm,
            global_vars={
                **kwargs
            },
            **kwargs,
        )
        return result


"""
AKR-Agent - A flexible rule-based AI Agent framework
"""

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "0.1.0.dev0"

from .agent import Agent
from .tools.base import Tool, ToolCenter

__all__ = ["Agent", "Tool", "ToolCenter"]



import loguru
from loguru._logger import Logger

logger = loguru.logger
import json
import os
import traceback
from pydantic import Field
from typing import AsyncGenerator
from ..libs.search.duckduckgo import DuckDuckGoSearchAPIWrapper
from ..libs.search.tavily import TavilySearchAPIWrapper
from .base import Tool


class DuckDuckGoSearchTool(Tool):

    name: str = "search_duckduckgo"
    description: str = (
        "A tool around Duck Duck Go Search. "
        "Useful for when you need to answer questions about current events. "
        "Input should be a search query. Output is a JSON array of the query results"
    )
    max_results: int = Field(alias="num_results", default=4)
    api_wrapper: DuckDuckGoSearchAPIWrapper = DuckDuckGoSearchAPIWrapper()
    backend: str = "text"

    async def run(self, keywords: str, **kwargs) -> AsyncGenerator[str, None]:
        """
        Run search for a query through DuckDuckGo engine and return results.

        Args:
            keywords: Most related keywords to search for.

        Returns:
            A list of dictionaries with the following keys:
                snippet - The description of the result.
                title - The title of the result.
                link - The link to the result.
        """
        yield self.api_wrapper.run(keywords=keywords)


class TavilySearchTool(Tool):
    name: str = "search_tavily"
    description: str = (
        "A wrapper around Tavily Search. "
        "Useful for when you need to answer questions about current events. "
        "Input should be a search query. Output is a JSON array of the query results"
    )
    max_results: int = 4
    api_wrapper: TavilySearchAPIWrapper = TavilySearchAPIWrapper(
        tavily_api_key=os.environ.get("TAVILY_API_KEY", None)
    )

    async def run(self, query: str, **kwargs) -> AsyncGenerator[str, None]:
        """
        Run search for a query through TavilySearch engine and return results.

        Args:
            query: The query to search for.
        """
        logger.info(f"调用 Tavily Search， query={query}")
        try:
            raw_results = await self.api_wrapper.raw_results_async(
                query=query, max_results=self.max_results
            )
        except Exception as e:
            logger.error("调用 Tavily Search 失败，", e, trace=traceback.format_exc())
            yield repr(e)
            return
        cleaned_results = self.api_wrapper.clean_results_with_images(raw_results)
        logger.info(
            "tavily search async, ",
            json.dumps(cleaned_results, indent=2, ensure_ascii=False),
            json.dumps(raw_results, indent=2, ensure_ascii=False),
        )
        yield cleaned_results
